"""The pinfo CLI tool."""

import argparse
import collections
import json
import os
import uuid

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.cli import logger
from plaso.cli import tool_options
from plaso.cli import tools
from plaso.cli import views
from plaso.cli.helpers import manager as helpers_manager
from plaso.containers import events
from plaso.containers import event_sources
from plaso.containers import reports
from plaso.containers import warnings
from plaso.engine import path_helper
from plaso.lib import errors
from plaso.lib import loggers
from plaso.serializer import json_serializer
from plaso.storage import factory as storage_factory


class PinfoTool(tools.CLITool, tool_options.StorageFileOptions):
  """Pinfo CLI tool.

  Attributes:
    compare_storage_information (bool): True if the tool is used to compare
        stores.
    generate_report (bool): True if a predefined report type should be
        generated.
    list_reports (bool): True if the report types should be listed.
    list_sections (bool): True if the section types should be listed.
  """

  NAME = 'pinfo'
  DESCRIPTION = (
      'Shows information about a Plaso storage file, for example how it was '
      'collected, what information was extracted from a source, etc.')

  _DEFAULT_HASH_TYPE = 'sha256'
  _HASH_CHOICES = ('md5', 'sha1', 'sha256')

  _REPORTS = {
      'browser_search': (
          'Report browser searches determined by the browser_search '
          'analysis plugin.'),
      'chrome_extension': (
          'Report Chrome extensions determined by the chrome_extension '
          'analysis plugin.'),
      'environment_variables': (
          'Report environment variables extracted during processing.'),
      'file_hashes': 'Report file hashes calculated during processing.',
      'windows_services': (
          'Report Windows services and drivers extracted during processing.'),
      'winevt_providers': (
          'Report Windows EventLog providers extracted during processing.')}

  _DEFAULT_REPORT_TYPE = 'none'
  _REPORT_CHOICES = sorted(list(_REPORTS.keys()) + ['list', 'none'])

  _SECTIONS = {
      'events': 'Show information about events.',
      'reports': 'Show information about analysis reports.',
      'sessions': 'Show information about sessions.',
      'sources': 'Show information about event sources.',
      'warnings': 'Show information about warnings during processing.'}

  _DEFAULT_OUTPUT_FORMAT = 'text'
  _SUPPORTED_OUTPUT_FORMATS = ('json', 'markdown', 'text')

  _CONTAINER_TYPE_ANALYSIS_REPORT = reports.AnalysisReport.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_DATA_STREAM = events.EventDataStream.CONTAINER_TYPE
  _CONTAINER_TYPE_EVENT_SOURCE = event_sources.EventSource.CONTAINER_TYPE
  _CONTAINER_TYPE_EXTRACTION_WARNING = warnings.ExtractionWarning.CONTAINER_TYPE
  _CONTAINER_TYPE_PREPROCESSING_WARNING = (
      warnings.PreprocessingWarning.CONTAINER_TYPE)
  _CONTAINER_TYPE_RECOVERY_WARNING = warnings.RecoveryWarning.CONTAINER_TYPE
  _CONTAINER_TYPE_TIMELINING_WARNING = warnings.TimeliningWarning.CONTAINER_TYPE

  def __init__(self, input_reader=None, output_writer=None):
    """Initializes the CLI tool object.

    Args:
      input_reader (Optional[InputReader]): input reader, where None indicates
          that the stdin input reader should be used.
      output_writer (Optional[OutputWriter]): output writer, where None
          indicates that the stdout output writer should be used.
    """
    super().__init__(
        input_reader=input_reader, output_writer=output_writer)
    self._compare_storage_file_path = None
    self._output_filename = None
    self._output_format = 'text'
    self._hash_type = self._DEFAULT_HASH_TYPE
    self._process_memory_limit = None
    self._report_type = self._DEFAULT_REPORT_TYPE
    self._sections = None
    self._storage_file_path = None
    self._verbose = False

    self.compare_storage_information = False
    self.generate_report = False
    self.list_reports = False
    self.list_sections = False

  def _CalculateStorageCounters(self, storage_reader):
    """Calculates the counters of the entire storage.

    Args:
      storage_reader (StorageReader): storage reader.

    Returns:
      dict[str, collections.Counter]: storage counters.
    """
    # TODO: determine analysis report counter from actual stored analysis
    # reports or remove.
    analysis_reports_counter = collections.Counter()
    analysis_reports_counter_error = False

    event_labels_counter = {}
    if storage_reader.HasAttributeContainers('event_label_count'):
      event_labels_counter = {
          event_label_count.label: event_label_count.number_of_events
          for event_label_count in storage_reader.GetAttributeContainers(
              'event_label_count')}

    event_labels_counter = collections.Counter(event_labels_counter)
    event_labels_counter_error = False

    parsers_counter = {}
    if storage_reader.HasAttributeContainers('parser_count'):
      parsers_counter = {
          parser_count.name: parser_count.number_of_events
          for parser_count in storage_reader.GetAttributeContainers(
              'parser_count')}

    parsers_counter = collections.Counter(parsers_counter)
    parsers_counter_error = False

    storage_counters = {}

    extraction_warnings_by_path_spec = collections.Counter()
    extraction_warnings_by_parser_chain = collections.Counter()

    for warning in storage_reader.GetAttributeContainers(
        self._CONTAINER_TYPE_EXTRACTION_WARNING):
      path_spec_string = self._GetPathSpecificationString(warning.path_spec)

      extraction_warnings_by_path_spec[path_spec_string] += 1
      extraction_warnings_by_parser_chain[warning.parser_chain] += 1

    storage_counters['extraction_warnings_by_path_spec'] = (
        extraction_warnings_by_path_spec)
    storage_counters['extraction_warnings_by_parser_chain'] = (
        extraction_warnings_by_parser_chain)

    recovery_warnings_by_path_spec = collections.Counter()
    recovery_warnings_by_parser_chain = collections.Counter()

    for warning in storage_reader.GetAttributeContainers(
        self._CONTAINER_TYPE_RECOVERY_WARNING):
      path_spec_string = self._GetPathSpecificationString(warning.path_spec)

      recovery_warnings_by_path_spec[path_spec_string] += 1
      recovery_warnings_by_parser_chain[warning.parser_chain] += 1

    storage_counters['recovery_warnings_by_path_spec'] = (
        recovery_warnings_by_path_spec)
    storage_counters['recovery_warnings_by_parser_chain'] = (
        recovery_warnings_by_parser_chain)

    timelining_warnings_by_path_spec = collections.Counter()
    timelining_warnings_by_parser_chain = collections.Counter()

    if storage_reader.HasAttributeContainers(
        self._CONTAINER_TYPE_TIMELINING_WARNING):
      for warning in storage_reader.GetAttributeContainers(
          self._CONTAINER_TYPE_TIMELINING_WARNING):
        path_spec_string = self._GetPathSpecificationString(warning.path_spec)

        timelining_warnings_by_path_spec[path_spec_string] += 1
        timelining_warnings_by_parser_chain[warning.parser_chain] += 1

    storage_counters['timelining_warnings_by_path_spec'] = (
        timelining_warnings_by_path_spec)
    storage_counters['timelining_warnings_by_parser_chain'] = (
        timelining_warnings_by_parser_chain)

    if not analysis_reports_counter_error:
      storage_counters['analysis_reports'] = analysis_reports_counter

    if not event_labels_counter_error:
      storage_counters['event_labels'] = event_labels_counter

    if not parsers_counter_error:
      storage_counters['parsers'] = parsers_counter

    return storage_counters

  def _CheckStorageFile(self, storage_file_path, warn_about_existing=False):
    """Checks if the storage file path is valid.

    Args:
      storage_file_path (str): path of the storage file.
      warn_about_existing (bool): True if the user should be warned about
          the storage file already existing.

    Raises:
      BadConfigOption: if the storage file path is invalid.
    """
    if os.path.exists(storage_file_path):
      if not os.path.isfile(storage_file_path):
        raise errors.BadConfigOption((
            f'Storage file: {storage_file_path:s} already exists and is not '
            f'a file.'))

      if warn_about_existing:
        logger.warning('Appending to an already existing storage file.')

    dirname = os.path.dirname(storage_file_path)
    if not dirname:
      dirname = '.'

    # TODO: add a more thorough check to see if the storage file really is
    # a plaso storage file.

    if not os.access(dirname, os.W_OK):
      raise errors.BadConfigOption(
          f'Unable to write to storage file: {storage_file_path:s}')

  def _CompareCounter(self, counter, compare_counter):
    """Compares two counters.

    Args:
      counter (collections.Counter): counter.
      compare_counter (collections.Counter): counter to compare against.

    Returns:
      dict[str, tuple[int, int]]: mismatching results per key.
    """
    keys = set(counter.keys())
    keys.union(compare_counter.keys())

    differences = {}
    for key in keys:
      value = counter.get(key, 0)
      compare_value = compare_counter.get(key, 0)
      if value != compare_value:
        differences[key] = (value, compare_value)

    return differences

  def _CompareStores(self, storage_reader, compare_storage_reader):
    """Compares the contents of two stores.

    Args:
      storage_reader (StorageReader): storage reader.
      compare_storage_reader (StorageReader): storage to compare against.

    Returns:
      bool: True if the content of the stores is identical.
    """
    stores_are_identical = True

    storage_counters = self._CalculateStorageCounters(storage_reader)
    compare_storage_counters = self._CalculateStorageCounters(
        compare_storage_reader)

    # Compare number of events.
    parsers_counter = storage_counters.get('parsers', collections.Counter())
    compare_parsers_counter = compare_storage_counters.get(
        'parsers', collections.Counter())
    differences = self._CompareCounter(parsers_counter, compare_parsers_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences,
          column_names=['Parser (plugin) name', 'Number of events'],
          title='Events generated per parser')

    # Compare extraction warnings by parser chain.
    warnings_counter = storage_counters.get(
        'extraction_warnings_by_parser_chain', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'extraction_warnings_by_parser_chain', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences,
          column_names=['Parser (plugin) name', 'Number of warnings'],
          title='Extraction warnings generated per parser')

    # Compare extraction warnings by path specification
    warnings_counter = storage_counters.get(
        'extraction_warnings_by_path_spec', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'extraction_warnings_by_path_spec', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Number of warnings', 'Pathspec'],
          reverse=True, title='Pathspecs with most extraction warnings')

    # Compare recovery warnings by parser chain.
    warnings_counter = storage_counters.get(
        'recovery_warnings_by_parser_chain', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'recovery_warnings_by_parser_chain', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences,
          column_names=['Parser (plugin) name', 'Number of warnings'],
          title='Recovery warnings generated per parser')

    # Compare recovery warnings by path specification
    warnings_counter = storage_counters.get(
        'recovery_warnings_by_path_spec', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'recovery_warnings_by_path_spec', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Number of warnings', 'Pathspec'],
          reverse=True, title='Pathspecs with most recovery warnings')

    # Compare timelining warnings by parser chain.
    warnings_counter = storage_counters.get(
        'timelining_warnings_by_parser_chain', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'timelining_warnings_by_parser_chain', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences,
          column_names=['Parser (plugin) name', 'Number of warnings'],
          title='Timelining warnings generated per parser')

    # Compare timelining warnings by path specification.
    warnings_counter = storage_counters.get(
        'timelining_warnings_by_path_spec', collections.Counter())
    compare_warnings_counter = compare_storage_counters.get(
        'timelining_warnings_by_path_spec', collections.Counter())
    differences = self._CompareCounter(
        warnings_counter, compare_warnings_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Number of warnings', 'Pathspec'],
          reverse=True, title='Pathspecs with most timelining warnings')

    # Compare event labels.
    labels_counter = storage_counters.get('event_labels', collections.Counter())
    compare_labels_counter = compare_storage_counters.get(
        'event_labels', collections.Counter())
    differences = self._CompareCounter(labels_counter, compare_labels_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Label', 'Number of event tags'],
          title='Event tags generated per label')

    # Compare analysis reports.
    reports_counter = storage_counters.get(
        'analysis_reports', collections.Counter())
    compare_reports_counter = compare_storage_counters.get(
        'analysis_reports', collections.Counter())
    differences = self._CompareCounter(reports_counter, compare_reports_counter)

    if differences:
      stores_are_identical = False

      self._PrintCounterDifferences(
          differences, column_names=['Plugin name', 'Number of reports'],
          title='Reports generated per plugin')

    return stores_are_identical

  def _GenerateAnalysisResultsReport(
      self, storage_reader, json_base_type, column_titles, container_type,
      attribute_names, attribute_mappings):
    """Generates an analysis results report.

    Args:
      storage_reader (StorageReader): storage reader.
      json_base_type (str): JSON base type.
      column_titles (list[str]): column titles of the Markdown and tab
          separated tables.
      container_type (str): attribute container type.
      attribute_names (list[str]): names of the attributes to report.
      attribute_mappings (dict[str, dict[st, str]]): mappings of attribute
          values to human readable strings.
    """
    self._GenerateReportHeader(json_base_type, column_titles)

    entry_format_string = self._GenerateReportEntryFormatString(attribute_names)

    if storage_reader.HasAttributeContainers(container_type):
      generator = storage_reader.GetAttributeContainers(container_type)

      for artifact_index, analysis_result in enumerate(generator):
        if self._output_format == 'json':
          if artifact_index > 0:
            self._output_writer.Write(',\n')

        attribute_values = analysis_result.CopyToDict()
        for key in attribute_names:
          value = attribute_values.get(key)
          if value is None:
            value = ''
          elif isinstance(value, int):
            value = attribute_mappings.get(key, {}).get(value, value)
          elif self._output_format == 'json':
            value = value.replace('\\', '\\\\')
          attribute_values[key] = value

        self._output_writer.Write(entry_format_string.format(
            **attribute_values))

    self._GenerateReportFooter()

  def _GenerateFileHashesReport(self, storage_reader):
    """Generates a file hashes report.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    hash_type = self._hash_type.upper()
    self._GenerateReportHeader(
        'file_hashes', [f'{hash_type:s} hash', 'Display name'])

    hash_attribute_name = f'{self._hash_type:s}_hash'
    entry_format_string = self._GenerateReportEntryFormatString([
        hash_attribute_name, 'display_name'])

    generator = storage_reader.GetAttributeContainers(
        self._CONTAINER_TYPE_EVENT_DATA_STREAM)

    for event_data_stream_index, event_data_stream in enumerate(generator):
      if self._output_format == 'json':
        if event_data_stream_index > 0:
          self._output_writer.Write(',\n')

      hash_value = getattr(
          event_data_stream, hash_attribute_name, None) or 'N/A'

      display_name = 'N/A'
      if event_data_stream.path_spec:
        display_name = path_helper.PathHelper.GetDisplayNameForPathSpec(
           event_data_stream.path_spec)

      attribute_values = {
          'display_name': display_name, hash_attribute_name: hash_value}

      self._output_writer.Write(entry_format_string.format(
          **attribute_values))

    self._GenerateReportFooter()

  def _GenerateReportEntryFormatString(self, attribute_names):
    """Generates a report entry format string.

    Args:
      attribute_names (list[str]): names of the attributes to report.

    Returns:
      str: entry format string.
    """
    if self._output_format == 'json':
      attributes_string = ', '.join([
          f'"{name:s}": "{{{name:s}!s}}"' for name in attribute_names])
      return f'    {{{{{attributes_string:s}}}}}'

    if self._output_format == 'markdown':
      attributes_string = ' | '.join([
          f'{{{name:s}!s}}' for name in attribute_names])
      return f'{attributes_string:s}\n'

    if self._output_format == 'text':
      attributes_string = '\t'.join([
          f'{{{name:s}!s}}' for name in attribute_names])
      return f'{attributes_string:s}\n'

    return ''

  def _GenerateReportFooter(self):
    """Generates a report footer."""
    if self._output_format == 'json':
      self._output_writer.Write('\n]}\n')

  def _GenerateReportHeader(self, json_base_type, column_titles):
    """Generates a report header.

    Args:
      json_base_type (str): JSON base type.
      column_titles (list[str]): column titles of the Markdown and tab
          separated tables.
    """
    if self._output_format == 'json':
      self._output_writer.Write(f'{{"{json_base_type:s}": [\n')

    elif self._output_format == 'markdown':
      titles_string = ' | '.join(column_titles)
      self._output_writer.Write(f'{titles_string:s}\n')
      separators_string = ' | '.join(['---'] * len(column_titles))
      self._output_writer.Write(f'{separators_string:s}\n')

    elif self._output_format == 'text':
      titles_string = '\t'.join(column_titles)
      self._output_writer.Write(f'{titles_string:s}\n')

  def _GenerateWinEvtProvidersReport(self, storage_reader):
    """Generates a Windows Event Log providers report.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    column_titles = [
        'Identifier', 'Log source(s)', 'Log type(s)', 'Event message file(s)',
        'Parameter message file(s)']
    self._GenerateReportHeader('winevt_providers', column_titles)

    attribute_names = [
        'identifier', 'log_sources', 'log_types', 'event_message_files',
        'parameter_message_files']
    entry_format_string = self._GenerateReportEntryFormatString(attribute_names)

    if storage_reader.HasAttributeContainers('windows_eventlog_provider'):
      generator = storage_reader.GetAttributeContainers(
          'windows_eventlog_provider')

      for artifact_index, artifact in enumerate(generator):
        if self._output_format == 'json':
          if artifact_index > 0:
            self._output_writer.Write(',\n')

        attribute_values = {
            'identifier': artifact.identifier or '',
            'event_message_files': artifact.event_message_files or [],
            'log_sources': artifact.log_sources or [],
            'log_types': artifact.log_types or [],
            'parameter_message_files': artifact.parameter_message_files or []}

        self._output_writer.Write(entry_format_string.format(
            **attribute_values))

    self._GenerateReportFooter()

  def _GetStorageReader(self, path):
    """Retrieves a storage reader.

    Args:
      path (str): path of the storage file.

    Returns:
      StorageReader: storage reader or None if the storage file format
          is not supported.

    Raises:
      BadConfigOption: if the storage file format is not supported.
    """
    storage_reader = (
        storage_factory.StorageFactory.CreateStorageReaderForFile(path))
    if not storage_reader:
      raise errors.BadConfigOption(
          f'Format of storage file: {path:s} not supported.')

    return storage_reader

  def _PrintAnalysisReportCounter(
      self, analysis_reports_counter, session_identifier=None):
    """Prints the analysis reports counter.

    Args:
      analysis_reports_counter (collections.Counter): number of analysis
          reports per analysis plugin.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if self._output_format == 'json':
      json_string = json.dumps(analysis_reports_counter)
      self._output_writer.Write(f'"analysis_reports": {json_string:s}')

    elif (self._output_format in ('markdown', 'text') and
          analysis_reports_counter):
      title = 'Reports generated per plugin'
      if session_identifier:
        title = ': '.join([title, session_identifier])

      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type,
          column_names=['Plugin name', 'Number of reports'], title=title)

      for key, value in sorted(analysis_reports_counter.items()):
        if key != 'total':
          table_view.AddRow([key, value])

      try:
        total = analysis_reports_counter['total']
      except KeyError:
        total = 'N/A'

      table_view.AddRow(['Total', total])

      table_view.Write(self._output_writer)

  def _PrintAnalysisReportsDetails(self, storage_reader):
    """Prints the details of the analysis reports.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    has_analysis_reports = storage_reader.HasAttributeContainers(
        self._CONTAINER_TYPE_ANALYSIS_REPORT)

    if self._output_format == 'text' and not has_analysis_reports:
      self._output_writer.Write('\nNo analysis reports stored.\n')

    else:
      generator = storage_reader.GetAttributeContainers(
          self._CONTAINER_TYPE_ANALYSIS_REPORT)

      for index, analysis_report in enumerate(generator):
        date_time_string = None
        if analysis_report.time_compiled is not None:
          date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
              timestamp=analysis_report.time_compiled)
          date_time_string = date_time.CopyToDateTimeStringISO8601()

        table_view = views.ViewsFactory.GetTableView(
            self._views_format_type, title=f'Analysis report: {index:d}')

        table_view.AddRow(['Name plugin', analysis_report.plugin_name or 'N/A'])
        table_view.AddRow(['Date and time', date_time_string or 'N/A'])
        table_view.AddRow([
            'Event filter', analysis_report.event_filter or 'N/A'])

        if not analysis_report.analysis_counter:
          table_view.AddRow(['Text', analysis_report.text or ''])
        else:
          table_view.AddRow(['Results', ''])
          for key, value in sorted(analysis_report.analysis_counter.items()):
            table_view.AddRow([key, value])

        table_view.Write(self._output_writer)

  def _PrintAnalysisReportSection(
      self, storage_reader, analysis_reports_counter):
    """Prints the analysis reports section.

    Args:
      storage_reader (StorageReader): storage reader.
      analysis_reports_counter (collections.Counter): number of analysis
          reports per analysis plugin.
    """
    self._PrintAnalysisReportCounter(analysis_reports_counter)

    if self._output_format in ('markdown', 'text'):
      self._PrintAnalysisReportsDetails(storage_reader)

  def _PrintCounterDifferences(
      self, differences, column_names=None, reverse=False, title=None):
    """Prints the counter differences.

    Args:
      differences (dict[str, tuple[int, int]]): mismatching results per key.
      column_names (Optional[list[str]]): column names.
      reverse (Optional[bool]): True if the key and values of differences
          should be printed in reverse order.
      title (Optional[str]): title.
    """
    # TODO: add support for 3 column table?
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=column_names, title=title)

    for key, value in sorted(differences.items()):
      value_string = f'{value[0]:d} ({value[1]:d})'
      if reverse:
        table_view.AddRow([value_string, key])
      else:
        table_view.AddRow([key, value_string])

    table_view.Write(self._output_writer)
    self._output_writer.Write('\n')

  def _PrintEventLabelsCounter(
      self, event_labels_counter, session_identifier=None):
    """Prints the event labels counter.

    Args:
      event_labels_counter (collections.Counter): number of event tags per
          label.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if self._output_format == 'json':
      json_string = json.dumps(event_labels_counter)
      self._output_writer.Write(f'"event_labels": {json_string:s}')

    elif self._output_format in ('markdown', 'text'):
      if self._output_format == 'text' and not event_labels_counter:
        if not session_identifier:
          self._output_writer.Write('\nNo events labels stored.\n')
      else:
        title = 'Event tags generated per label'

        if session_identifier:
          title = ': '.join([title, session_identifier])
          title_level = 4
        else:
          title_level = 2

        if not event_labels_counter:
          if not session_identifier:
            heading_marker = '#' * title_level
            self._output_writer.Write(
                f'{heading_marker:s} {title:s}\n\nN/A\n\n')
        else:
          table_view = views.ViewsFactory.GetTableView(
              self._views_format_type,
              column_names=['Label', 'Number of event tags'], title=title,
              title_level=title_level)

          for key, value in sorted(event_labels_counter.items()):
            if key != 'total':
              table_view.AddRow([key, value])

          try:
            total = event_labels_counter['total']
          except KeyError:
            total = 'N/A'

          table_view.AddRow(['Total', total])

          table_view.Write(self._output_writer)

  def _PrintParsersCounter(self, parsers_counter, session_identifier=None):
    """Prints the parsers counter.

    Args:
      parsers_counter (collections.Counter): number of events per parser or
          parser plugin.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if self._output_format == 'json':
      if session_identifier:
        self._output_writer.Write(', ')

      json_string = json.dumps(parsers_counter)
      self._output_writer.Write(f'"parsers": {json_string:s}')

    elif self._output_format in ('markdown', 'text'):
      if self._output_format == 'text' and not parsers_counter:
        if not session_identifier:
          self._output_writer.Write('\nNo events stored.\n')

      else:
        title = 'Events generated per parser'
        if session_identifier:
          title = ': '.join([title, session_identifier])
          title_level = 4
        else:
          title_level = 2

        if not parsers_counter:
          if not session_identifier:
            heading_marker = '#' * title_level
            self._output_writer.Write(
                f'{heading_marker:s} {title:s}\n\nN/A\n\n')
        else:
          table_view = views.ViewsFactory.GetTableView(
              self._views_format_type,
              column_names=['Parser (plugin) name', 'Number of events'],
              title=title, title_level=title_level)

          for key, value in sorted(parsers_counter.items()):
            if key != 'total':
              table_view.AddRow([key, value])

          table_view.AddRow(['Total', parsers_counter['total']])

          table_view.Write(self._output_writer)

  def _PrintPreprocessingWarningsDetails(self, storage_reader):
    """Prints the details of the preprocessing warnings.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    generator = storage_reader.GetAttributeContainers(
        self._CONTAINER_TYPE_PREPROCESSING_WARNING)

    for index, warning in enumerate(generator):
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, title=f'Preprocessing warning: {index:d}')

      table_view.AddRow(['Message', warning.message])
      table_view.AddRow(['Plugin name', warning.plugin_name])

      if warning.path_spec:
        # TODO: add helper method to format path spec as row.
        path_spec_string = self._GetPathSpecificationString(warning.path_spec)

        for path_index, line in enumerate(path_spec_string.split('\n')):
          if not line:
            continue

          if path_index == 0:
            table_view.AddRow(['Path specification', line])
          else:
            table_view.AddRow(['', line])

      table_view.Write(self._output_writer)

  def _PrintWarningsDetails(self, storage_reader, container_type, warning_type):
    """Prints the details of warnings.

    Args:
      storage_reader (StorageReader): storage reader.
      container_type (str): attribute container type.
      warning_type (str): warning type.
    """
    generator = storage_reader.GetAttributeContainers(container_type)

    for index, warning in enumerate(generator):
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, title=f'{warning_type:s} warning: {index:d}')

      table_view.AddRow(['Message', warning.message])
      table_view.AddRow(['Parser chain', warning.parser_chain])

      path_spec_string = self._GetPathSpecificationString(warning.path_spec)

      for path_index, line in enumerate(path_spec_string.split('\n')):
        if not line:
          continue

        if path_index == 0:
          table_view.AddRow(['Path specification', line])
        else:
          table_view.AddRow(['', line])

      table_view.Write(self._output_writer)

  def _PrintSessionDetailsAsJSON(self, session):
    """Prints the details of a session as JSON.

    Args:
      session (Session): session.
    """
    json_string = (
        json_serializer.JSONAttributeContainerSerializer.WriteSerialized(
            session))
    self._output_writer.Write(f'"session": {json_string:s}')

  def _PrintSessionDetailsAsTable(self, session, session_identifier):
    """Prints the details of a session as a table.

    Args:
      session (Session): session.
      session_identifier (str): session identifier, formatted as a UUID.
    """
    start_time = 'N/A'
    if session.start_time is not None:
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=session.start_time)
      start_time = date_time.CopyToDateTimeStringISO8601()

    completion_time = 'N/A'
    if session.completion_time is not None:
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=session.completion_time)
      completion_time = date_time.CopyToDateTimeStringISO8601()

    enabled_parser_names = 'N/A'
    if session.enabled_parser_names:
      enabled_parser_names = ', '.join(sorted(session.enabled_parser_names))

    command_line_arguments = session.command_line_arguments or 'N/A'
    parser_filter_expression = session.parser_filter_expression or 'N/A'
    preferred_encoding = session.preferred_encoding or 'N/A'
    preferred_time_zone = session.preferred_time_zone or 'N/A'

    if session.artifact_filters:
      artifact_filters_string = ', '.join(session.artifact_filters)
    else:
      artifact_filters_string = 'N/A'
    filter_file = session.filter_file or 'N/A'

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title=f'Session: {session_identifier:s}')

    table_view.AddRow(['Start time', start_time])
    table_view.AddRow(['Completion time', completion_time])
    table_view.AddRow(['Product name', session.product_name])
    table_view.AddRow(['Product version', session.product_version])
    table_view.AddRow(['Command line arguments', command_line_arguments])
    table_view.AddRow(['Parser filter expression', parser_filter_expression])
    table_view.AddRow(['Enabled parser and plugins', enabled_parser_names])
    table_view.AddRow(['Preferred encoding', preferred_encoding])
    table_view.AddRow(['Preferred time zone', preferred_time_zone])
    table_view.AddRow(['Debug mode', session.debug_mode])
    table_view.AddRow(['Artifact filters', artifact_filters_string])
    table_view.AddRow(['Filter file', filter_file])

    table_view.Write(self._output_writer)

  def _PrintSessionsDetails(self, storage_reader):
    """Prints the details of the sessions.

    Args:
      storage_reader (BaseStore): storage.
    """
    if self._output_format == 'json':
      self._output_writer.Write('"sessions": {')

    for session_index, session in enumerate(storage_reader.GetSessions()):
      session_identifier = str(uuid.UUID(hex=session.identifier))

      if self._output_format == 'json':
        if session_index != 0:
          self._output_writer.Write(', ')

        self._PrintSessionDetailsAsJSON(session)

      elif self._output_format in ('markdown', 'text'):
        self._PrintSessionDetailsAsTable(session, session_identifier)

      if self._verbose:
        system_configuration = storage_reader.GetAttributeContainerByIndex(
            'system_configuration', session_index)
        if system_configuration:
          self._PrintSystemConfigurations(
              storage_reader, [system_configuration],
              session_identifier=session_identifier)

    if self._output_format == 'json':
      self._output_writer.Write('}')

  def _PrintSessionsOverview(self, storage_reader):
    """Prints a sessions overview.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title='Sessions', title_level=2)

    for session in storage_reader.GetSessions():
      date_time = dfdatetime_posix_time.PosixTimeInMicroseconds(
          timestamp=session.start_time)
      start_time = date_time.CopyToDateTimeStringISO8601()

      session_identifier = str(uuid.UUID(hex=session.identifier))
      table_view.AddRow([session_identifier, start_time])

    table_view.Write(self._output_writer)

  def _PrintSessionsSection(self, storage_reader):
    """Prints the session section.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    if self._output_format in ('markdown', 'text'):
      self._PrintSessionsOverview(storage_reader)

    if (self._output_format == 'json' or self._verbose or
        'sessions' in self._sections):
      self._PrintSessionsDetails(storage_reader)

  def _PrintSystemConfiguration(
      self, storage_reader, system_configuration, session_identifier=None):
    """Prints the details of a system configuration.

    Args:
      storage_reader (StorageReader): storage reader.
      system_configuration (SystemConfiguration): system configuration.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if not system_configuration:
      return

    title = 'System configuration'
    if session_identifier:
      title = ': '.join([title, session_identifier])

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title=title)

    hostname = 'N/A'
    if system_configuration.hostname:
      hostname = system_configuration.hostname.name

    operating_system = system_configuration.operating_system or 'N/A'
    operating_system_product = (
        system_configuration.operating_system_product or 'N/A')
    operating_system_version = (
        system_configuration.operating_system_version or 'N/A')
    language = system_configuration.language or 'N/A'
    code_page = system_configuration.code_page or 'N/A'
    keyboard_layout = system_configuration.keyboard_layout or 'N/A'
    time_zone = system_configuration.time_zone or 'N/A'

    table_view.AddRow(['Hostname', hostname])
    table_view.AddRow(['Operating system', operating_system])
    table_view.AddRow(['Operating system product', operating_system_product])
    table_view.AddRow(['Operating system version', operating_system_version])
    table_view.AddRow(['Language', language])
    table_view.AddRow(['Code page', code_page])
    table_view.AddRow(['Keyboard layout', keyboard_layout])
    table_view.AddRow(['Time zone', time_zone])

    table_view.Write(self._output_writer)

    title = 'Available time zones'
    if session_identifier:
      title = ': '.join([title, session_identifier])

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Name', 'Offset from UTC'], title=title)

    # TODO: filter time zones on specific system configuration.
    for time_zone in storage_reader.GetAttributeContainers('time_zone'):
      hours_from_utc, minutes_from_utc = divmod(time_zone.offset, 60)
      if hours_from_utc < 0:
        sign = '+'
        hours_from_utc *= -1
      else:
        sign = '-'

      time_zone_offset = f'{sign:s}{hours_from_utc:02d}:{minutes_from_utc:02d}'
      table_view.AddRow([time_zone.name, time_zone_offset])

    table_view.Write(self._output_writer)

    title = 'User accounts'
    if session_identifier:
      title = ': '.join([title, session_identifier])

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type,
        column_names=['Username', 'User directory'], title=title)

    # TODO: filter user accounts on specific system configuration.
    for user_account in storage_reader.GetAttributeContainers('user_account'):
      table_view.AddRow([user_account.username, user_account.user_directory])

    table_view.Write(self._output_writer)

  def _PrintSystemConfigurations(
      self, storage_reader, system_configurations, session_identifier=None):
    """Prints the details of system configurations.

    Args:
      storage_reader (StorageReader): storage reader.
      system_configurations (list[SystemConfiguration]): system configurations.
      session_identifier (Optional[str]): session identifier, formatted as
          a UUID.
    """
    if self._output_format == 'json':
      self._output_writer.Write(', "system_configurations": {')

    for configuration_index, configuration in enumerate(system_configurations):
      if self._output_format == 'json':
        if configuration_index != 0:
          self._output_writer.Write(', ')

        json_string = (
            json_serializer.JSONAttributeContainerSerializer.WriteSerialized(
                configuration.system_configuration))
        self._output_writer.Write(f'"system_configuration": {json_string:s}')

      elif self._output_format in ('markdown', 'text'):
        self._PrintSystemConfiguration(
            storage_reader, configuration,
            session_identifier=session_identifier)

    if self._output_format == 'json':
      self._output_writer.Write('}')

  def _PrintStorageInformation(self, storage_reader):
    """Prints information about the store.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    if self._output_format == 'json':
      self._output_writer.Write('{')
    elif self._output_format in ('markdown', 'text'):
      self._PrintStorageOverviewAsTable(storage_reader)

    section_written = False

    if self._sections == 'all' or 'sessions' in self._sections:
      self._PrintSessionsSection(storage_reader)
      section_written = True

    if self._sections == 'all' or 'sources' in self._sections:
      if self._output_format == 'json' and section_written:
        self._output_writer.Write(', ')

      self._PrintSourcesOverview(storage_reader)
      section_written = True

    if self._output_format == 'json' and section_written:
      self._output_writer.Write(', ')

    storage_counters = self._CalculateStorageCounters(storage_reader)

    if self._output_format == 'json':
      self._output_writer.Write('"storage_counters": {')

    section_written = False

    if self._sections == 'all' or 'events' in self._sections:
      parsers = storage_counters.get('parsers', collections.Counter())

      self._PrintParsersCounter(parsers)
      section_written = True

      event_labels = storage_counters.get(
          'event_labels', collections.Counter())

      if self._output_format == 'json' and section_written:
        self._output_writer.Write(', ')

      self._PrintEventLabelsCounter(event_labels)

    if self._sections == 'all' or 'warnings' in self._sections:
      if self._output_format == 'json' and section_written:
        self._output_writer.Write(', ')

      self._PrintWarningsSection(storage_reader, storage_counters)
      section_written = True

    if self._sections == 'all' or 'reports' in self._sections:
      if self._output_format == 'json' and section_written:
        self._output_writer.Write(', ')

      analysis_reports = storage_counters.get(
          'analysis_reports', collections.Counter())

      self._PrintAnalysisReportSection(storage_reader, analysis_reports)

    if self._output_format == 'json':
      self._output_writer.Write('}')

    if self._output_format == 'json':
      self._output_writer.Write('}')
    elif self._output_format in ('markdown', 'text'):
      self._output_writer.Write('\n')

  def _PrintSourcesOverview(self, storage_reader):
    """Prints a sources overview.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    if self._output_format == 'json':
      self._output_writer.Write('"event_sources": {')

    elif self._output_format in ('markdown', 'text'):
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type, title='Event sources', title_level=2)

    generator = storage_reader.GetAttributeContainers(
        self._CONTAINER_TYPE_EVENT_SOURCE)

    number_of_event_sources = 0
    for source_index, source in enumerate(generator):
      if self._output_format == 'json':
        if source_index != 0:
          self._output_writer.Write(', ')

        json_string = (
            json_serializer.JSONAttributeContainerSerializer.WriteSerialized(
                source))
        self._output_writer.Write(f'"source": {json_string:s}')

      elif self._output_format in ('markdown', 'text'):
        if self._verbose or 'sources' in self._sections:
          path_spec_string = self._GetPathSpecificationString(source.path_spec)

          for path_index, line in enumerate(path_spec_string.split('\n')):
            if not line:
              continue

            if path_index == 0:
              table_view.AddRow([f'{source_index:d}', line])
            else:
              table_view.AddRow(['', line])

      number_of_event_sources += 1

    if self._output_format == 'json':
      self._output_writer.Write('}')
    elif self._output_format in ('markdown', 'text'):
      if not (self._verbose or 'sources' in self._sections):
        table_view.AddRow(['Total', f'{number_of_event_sources:d}'])
      table_view.Write(self._output_writer)

  def _PrintStorageOverviewAsTable(self, storage_reader):
    """Prints a storage overview as a table.

    Args:
      storage_reader (StorageReader): storage reader.
    """
    format_version = storage_reader.GetFormatVersion()
    serialization_format = storage_reader.GetSerializationFormat()

    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, title='Plaso Storage Information',
        title_level=1)
    table_view.AddRow(['Filename', os.path.basename(self._storage_file_path)])
    table_view.AddRow(['Format version', format_version])
    table_view.AddRow(['Serialization format', serialization_format])
    table_view.Write(self._output_writer)

  def _PrintWarningCountersJSON(
      self, warnings_by_path_spec, warnings_by_parser_chain):
    """Prints JSON containing a summary of the number of warnings.

    Args:
      warnings_by_path_spec (collections.Counter): number of warnings per
          path specification.
      warnings_by_parser_chain (collections.Counter): number of warnings per
          parser chain.
    """
    json_string = json.dumps(warnings_by_parser_chain)
    self._output_writer.Write(f'"warnings_by_parser": {json_string:s}')

    json_string = json.dumps(warnings_by_path_spec)
    self._output_writer.Write(f', "warnings_by_path_spec": {json_string:s}')

  def _PrintWarningCountersTable(
      self, description, warnings_by_path_spec, warnings_by_parser_chain):
    """Prints a table containing a summary of the number of warnings.

    Args:
      description (str): description of the type of warning.
      warnings_by_path_spec (collections.Counter): number of warnings per
          path specification.
      warnings_by_parser_chain (collections.Counter): number of warnings per
          parser chain.
    """
    if warnings_by_parser_chain:
      title = description.title()
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type,
          column_names=['Parser (plugin) name', 'Number of warnings'],
          title=f'{title:s} warnings generated per parser')
      for parser_chain, count in warnings_by_parser_chain.items():
        parser_chain = parser_chain or '<No parser>'
        table_view.AddRow([parser_chain, f'{count:d}'])
      table_view.Write(self._output_writer)

    if warnings_by_path_spec:
      table_view = views.ViewsFactory.GetTableView(
          self._views_format_type,
          column_names=['Number of warnings', 'Pathspec'],
          title=f'Path specifications with most {description:s} warnings')

      for path_spec, count in warnings_by_path_spec.most_common(10):
        for path_index, line in enumerate(path_spec.split('\n')):
          if not line:
            continue

          if path_index == 0:
            table_view.AddRow([f'{count:d}', line])
          else:
            table_view.AddRow(['', line])

      table_view.Write(self._output_writer)

  def _PrintWarningsSection(self, storage_reader, storage_counters):
    """Prints the warnings section.

    Args:
      storage_reader (StorageReader): storage reader.
      storage_counters (dict[str, collections.Counter]): storage counters.
    """
    has_extraction_warnings = storage_reader.HasAttributeContainers(
        self._CONTAINER_TYPE_EXTRACTION_WARNING)

    has_preprocessing_warnings = storage_reader.HasAttributeContainers(
        self._CONTAINER_TYPE_PREPROCESSING_WARNING)

    has_recovery_warnings = storage_reader.HasAttributeContainers(
        self._CONTAINER_TYPE_RECOVERY_WARNING)

    if (self._output_format == 'text' and not has_extraction_warnings and
        not has_preprocessing_warnings and not has_recovery_warnings):
      self._output_writer.Write('\nNo warnings stored.\n')

    else:
      warnings_by_path_spec = storage_counters.get(
          'extraction_warnings_by_path_spec', collections.Counter())
      warnings_by_parser_chain = storage_counters.get(
          'extraction_warnings_by_parser_chain', collections.Counter())

      if self._output_format == 'json':
        self._PrintWarningCountersJSON(
            warnings_by_path_spec, warnings_by_parser_chain)

      elif self._output_format in ('markdown', 'text'):
        self._PrintWarningCountersTable(
            'extraction', warnings_by_path_spec, warnings_by_parser_chain)

        if self._verbose or 'warnings' in self._sections:
          self._PrintWarningsDetails(
              storage_reader, self._CONTAINER_TYPE_EXTRACTION_WARNING,
              'Extraction')

      warnings_by_path_spec = storage_counters.get(
          'preprocessing_warnings_by_path_spec', collections.Counter())
      warnings_by_parser_chain = storage_counters.get(
          'preprocessing_warnings_by_parser_chain', collections.Counter())

      # TODO: print preprocessing warnings as part of JSON output format.

      if self._output_format in ('markdown', 'text'):
        self._PrintWarningCountersTable(
            'preprocessing', warnings_by_path_spec, warnings_by_parser_chain)

        if self._verbose or 'warnings' in self._sections:
          self._PrintPreprocessingWarningsDetails(storage_reader)

      warnings_by_path_spec = storage_counters.get(
          'recovery_warnings_by_path_spec', collections.Counter())
      warnings_by_parser_chain = storage_counters.get(
          'recovery_warnings_by_parser_chain', collections.Counter())

      # TODO: print recovery warnings as part of JSON output format.

      if self._output_format in ('markdown', 'text'):
        self._PrintWarningCountersTable(
            'recovery', warnings_by_path_spec, warnings_by_parser_chain)

        if self._verbose or 'warnings' in self._sections:
          self._PrintWarningsDetails(
              storage_reader, self._CONTAINER_TYPE_RECOVERY_WARNING,
              'Recovery')

      warnings_by_path_spec = storage_counters.get(
          'timelining_warnings_by_path_spec', collections.Counter())
      warnings_by_parser_chain = storage_counters.get(
          'timelining_warnings_by_parser_chain', collections.Counter())

      # TODO: print timelining warnings as part of JSON output format.

      if self._output_format in ('markdown', 'text'):
        self._PrintWarningCountersTable(
            'timelining', warnings_by_path_spec, warnings_by_parser_chain)

        if self._verbose or 'warnings' in self._sections:
          self._PrintWarningsDetails(
              storage_reader, self._CONTAINER_TYPE_TIMELINING_WARNING,
              'Timelining')

  def CompareStores(self):
    """Compares the contents of two stores.

    Returns:
      bool: True if the content of the stores is identical.

    Raises:
      BadConfigOption: if the storage file format is not supported.
    """
    storage_reader = self._GetStorageReader(self._storage_file_path)
    compare_storage_reader = self._GetStorageReader(
        self._compare_storage_file_path)

    try:
      result = self._CompareStores(storage_reader, compare_storage_reader)

    finally:
      compare_storage_reader.Close()
      storage_reader.Close()

    if result:
      self._output_writer.Write('Storage files are identical.\n')
    else:
      self._output_writer.Write('Storage files are different.\n')

    return result

  def GenerateReport(self):
    """Generates a report.

    Raises:
      BadConfigOption: if the storage file format is not supported.
    """
    storage_reader = self._GetStorageReader(self._storage_file_path)

    try:
      if self._report_type == 'browser_search':
        column_titles = ['Search engine', 'Search term', 'Number of queries']
        attribute_names = ['search_engine', 'search_term', 'number_of_queries']
        attribute_mappings = {}
        self._GenerateAnalysisResultsReport(
            storage_reader, 'browser_searches', column_titles,
            'browser_search_analysis_result', attribute_names,
            attribute_mappings)

      elif self._report_type == 'chrome_extension':
        column_titles = ['Username', 'Extension identifier', 'Extension']
        attribute_names = ['username', 'extension_identifier', 'extension']
        attribute_mappings = {}
        self._GenerateAnalysisResultsReport(
            storage_reader, 'chrome_extensions', column_titles,
            'chrome_extension_analysis_result', attribute_names,
            attribute_mappings)

      elif self._report_type == 'environment_variables':
        column_titles = ['Name', 'Value']
        attribute_names = ['name', 'value']
        attribute_mappings = {}
        self._GenerateAnalysisResultsReport(
            storage_reader, 'environment_variables', column_titles,
            'environment_variable', attribute_names, attribute_mappings)

      elif self._report_type == 'file_hashes':
        self._GenerateFileHashesReport(storage_reader)

      elif self._report_type == 'windows_services':
        column_titles = ['Name', 'Service type', 'Start type', 'Image path']
        attribute_names = ['name', 'service_type', 'start_type', 'image_path']
        # TODO: consider using message formatting helpers?
        attribute_mappings = {
            'service_type': {
                0x01: 'Kernel device driver (0x01)',
                0x02: 'File system driver (0x02)',
                0x04: 'Adapter (0x04)',
                0x10: 'Stand-alone service (0x10)',
                0x20: 'Shared service (0x20)'},
            'start_type': {
                0: 'Boot (0)',
                1: 'System (1)',
                2: 'Automatic (2)',
                3: 'On demand (3)',
                4: 'Disabled (4)'}
        }
        self._GenerateAnalysisResultsReport(
            storage_reader, 'windows_services', column_titles,
            'windows_service_configuration', attribute_names,
            attribute_mappings)

      elif self._report_type == 'winevt_providers':
        self._GenerateWinEvtProvidersReport(storage_reader)

    finally:
      storage_reader.Close()

  def ListReports(self):
    """Lists information about the available report types."""
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Reports')

    for name, description in sorted(self._REPORTS.items()):
      table_view.AddRow([name, description])
    table_view.Write(self._output_writer)

  def ListSections(self):
    """Lists information about the available sections."""
    table_view = views.ViewsFactory.GetTableView(
        self._views_format_type, column_names=['Name', 'Description'],
        title='Sections')

    for name, description in sorted(self._SECTIONS.items()):
      table_view.AddRow([name, description])
    table_view.Write(self._output_writer)

  def ParseArguments(self, arguments):
    """Parses the command line arguments.

    Args:
      arguments (list[str]): command line arguments.

    Returns:
      bool: True if the arguments were successfully parsed.
    """
    loggers.ConfigureLogging()

    argument_parser = argparse.ArgumentParser(
        description=self.DESCRIPTION, add_help=False,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    self.AddBasicOptions(argument_parser)
    self.AddStorageOptions(argument_parser)
    self.AddLogFileOptions(argument_parser)

    if self._CanEnforceProcessMemoryLimit():
      helpers_manager.ArgumentHelperManager.AddCommandLineArguments(
          argument_parser, names=['process_resources'])

    argument_parser.add_argument(
        '--compare', dest='compare_storage_file', type=str,
        action='store', default='', metavar='STORAGE_FILE', help=(
            'The path of the storage file to compare against.'))

    output_formats = ', '.join(sorted(self._SUPPORTED_OUTPUT_FORMATS))
    argument_parser.add_argument(
        '--output_format', '--output-format', dest='output_format', type=str,
        choices=self._SUPPORTED_OUTPUT_FORMATS, action='store',
        default=self._DEFAULT_OUTPUT_FORMAT, metavar='FORMAT', help=(
            f'Format of the output, the default is: '
            f'{self._DEFAULT_OUTPUT_FORMAT:s}. Supported options: '
            f'{output_formats:s}.'))

    hash_choices = ', '.join(self._HASH_CHOICES)
    argument_parser.add_argument(
        '--hash', dest='hash', choices=self._HASH_CHOICES, action='store',
        metavar='TYPE', default=self._DEFAULT_HASH_TYPE, help=(
            f'Type of hash to output in file_hashes report. Supported options: '
            f'{hash_choices:s}'))

    report_choices = ', '.join(self._REPORT_CHOICES)
    argument_parser.add_argument(
        '--report', dest='report', choices=self._REPORT_CHOICES, action='store',
        metavar='TYPE', default=self._DEFAULT_REPORT_TYPE, help=(
            f'Report on specific information. Supported options: '
            f'{report_choices:s}'))

    argument_parser.add_argument(
        '--sections', dest='sections', type=str, action='store', default='all',
        metavar='SECTIONS_LIST', help=(
            'List of sections to output. This is a comma separated list where '
            'each entry is the name of a section. Use "--sections list" to '
            'list the available sections and "--sections all" to show all '
            'available sections.'))

    argument_parser.add_argument(
        '-v', '--verbose', dest='verbose', action='store_true',
        default=False, help='Print verbose output.')

    argument_parser.add_argument(
        '-w', '--write', metavar='OUTPUTFILE', dest='write',
        help='Output filename.')

    try:
      options = argument_parser.parse_args(arguments)
    except UnicodeEncodeError:
      # If we get here we are attempting to print help in a non-Unicode
      # terminal.
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_help())
      return False

    try:
      self.ParseOptions(options)
    except errors.BadConfigOption as exception:
      self._output_writer.Write(f'ERROR: {exception!s}\n')
      self._output_writer.Write('\n')
      self._output_writer.Write(argument_parser.format_usage())
      return False

    self._WaitUserWarning()

    loggers.ConfigureLogging(
        debug_output=self._debug_mode, filename=self._log_file,
        quiet_mode=self._quiet_mode)

    return True

  def ParseOptions(self, options):
    """Parses the options.

    Args:
      options (argparse.Namespace): command line arguments.

    Raises:
      BadConfigOption: if the options are invalid.
    """
    self._ParseInformationalOptions(options)

    self._ParseLogFileOptions(options)

    self._verbose = getattr(options, 'verbose', False)

    self._report_type = getattr(
        options, 'report', self._DEFAULT_REPORT_TYPE)
    self._sections = getattr(options, 'sections', '')

    self.list_reports = self._report_type == 'list'
    self.list_sections = self._sections == 'list'

    self.show_troubleshooting = getattr(options, 'show_troubleshooting', False)
    if self.list_reports or self.list_sections or self.show_troubleshooting:
      return

    if self._report_type != 'none':
      if self._report_type not in self._REPORTS:
        raise errors.BadConfigOption(
            f'Unsupported report type: {self._report_type:s}')

      self.generate_report = True

    if self._sections != 'all':
      self._sections = self._sections.split(',')

    self._hash_type = getattr(options, 'hash', self._DEFAULT_HASH_TYPE)
    self._output_filename = getattr(options, 'write', None)

    helpers_manager.ArgumentHelperManager.ParseOptions(
        options, self, names=['process_resources'])

    # TODO: move check into _CheckStorageFile.
    self._storage_file_path = self.ParseStringOption(options, 'storage_file')
    if not self._storage_file_path:
      raise errors.BadConfigOption('Missing storage file option.')

    if not os.path.isfile(self._storage_file_path):
      raise errors.BadConfigOption(
          f'No such storage file: {self._storage_file_path:s}')

    compare_storage_file_path = self.ParseStringOption(
        options, 'compare_storage_file')
    if compare_storage_file_path:
      if not os.path.isfile(compare_storage_file_path):
        raise errors.BadConfigOption(
            f'No such storage file: {compare_storage_file_path:s}')

      self._compare_storage_file_path = compare_storage_file_path
      self.compare_storage_information = True

    self._output_format = self.ParseStringOption(options, 'output_format')

    if self._output_filename:
      if os.path.exists(self._output_filename):
        raise errors.BadConfigOption(
            f'Output file: {self._output_filename:s} already exists.')
      output_file_object = open(self._output_filename, 'wb')  # pylint: disable=consider-using-with
      self._output_writer = tools.FileObjectOutputWriter(output_file_object)

    self._EnforceProcessMemoryLimit(self._process_memory_limit)

  def PrintStorageInformation(self):
    """Prints the storage information.

    Raises:
      BadConfigOption: if the storage file format is not supported.
    """
    if self._output_format == 'markdown':
      self._views_format_type = views.ViewsFactory.FORMAT_TYPE_MARKDOWN
    elif self._output_format == 'text':
      self._views_format_type = views.ViewsFactory.FORMAT_TYPE_CLI

    storage_reader = self._GetStorageReader(self._storage_file_path)

    try:
      self._PrintStorageInformation(storage_reader)
    finally:
      storage_reader.Close()



"""Plist parser plugin for iOS identity services status cache files."""

from dfdatetime import cocoa_time as dfdatetime_cocoa_time

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class IOSIdstatusacheEventData(events.EventData):
  """iOS identity services status cache event data.

  Attributes:
    apple_identifier (str): type and value of the identifier.
    lookup_time (dfdatetime.DateTimeValues): date and time of the lookup.
    process_name (str)" name of the process that looked up an identifier.
  """

  DATA_TYPE = 'ios:idstatuscache:lookup'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.apple_identifier = None
    self.lookup_time = None
    self.process_name = None


class IOSIdstatusachePlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for identity services status cache files.

  Identity services status cache plist files are typically named:
  com.apple.identityservices.idstatuscache.plist
  """

  NAME = 'ios_identityservices'
  DATA_FORMAT = 'Idstatuscache plist file'

  PLIST_PATH_FILTERS = frozenset([interface.PlistPathFilter(
      'com.apple.identityservices.idstatuscache.plist')])

  def _GetDateTimeValueFromPlistKey(self, plist_key, plist_value_name):
    """Retrieves a date and time value from a specific value in a plist key.

    Args:
      plist_key (object): plist key.
      plist_value_name (str): name of the value in the plist key.

    Returns:
      dfdatetime.TimeElementsInMicroseconds: date and time or None if not
          available.
    """
    timestamp = plist_key.get(plist_value_name)
    if timestamp is None:
      return None

    return dfdatetime_cocoa_time.CocoaTime(timestamp=timestamp)

  # pylint: disable=arguments-differ
  def _ParsePlist(
      self, parser_mediator, match=None, top_level=None, **unused_kwargs):
    """Extracts identity services status cache information from the plist.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
      top_level (Optional[dict[str, object]]): plist top-level item.
    """
    for _, process_name, process_values in self._RecurseKey(top_level, depth=1):
      if process_name == 'CacheVersion':
        continue

      for apple_identifier, apple_identifier_values in process_values.items():
        event_data = IOSIdstatusacheEventData()
        event_data.apple_identifier = apple_identifier
        event_data.lookup_time = self._GetDateTimeValueFromPlistKey(
            apple_identifier_values, 'LookupDate')
        event_data.process_name = process_name

        parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(IOSIdstatusachePlistPlugin)



"""JSON-L parser plugin for Docker layer configuration files."""

from plaso.containers import events
from plaso.parsers import jsonl_parser
from plaso.parsers.jsonl_plugins import interface


class DockerLayerConfigurationEventData(events.EventData):
  """Docker layer configuration event data.

  Attributes:
    command: the command used which made Docker create a new layer.
    creation_time (dfdatetime.DateTimeValues): date and time the layer
        was created (added).
    layer_identifier: the identifier of the current Docker layer (SHA-1).
  """

  DATA_TYPE = 'docker:layer:configuration'

  def __init__(self):
    """Initializes event data."""
    super().__init__(
        data_type=self.DATA_TYPE)
    self.command = None
    self.creation_time = None
    self.layer_identifier = None


class DockerLayerConfigurationJSONLPlugin(interface.JSONLPlugin):
  """JSON-L parser plugin for Docker layer configuration files.

  This parser handles per Docker layer configuration files stored in:
  DOCKER_DIR/graph/<layer_identifier>/json
  """

  NAME = 'docker_layer_config'
  DATA_FORMAT = 'Docker layer configuration file'

  def _GetLayerIdentifierFromPath(self, parser_mediator):
    """Extracts a layer (or graph) identifier from a path.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.

    Returns:
      str: layer identifier.
    """
    file_entry = parser_mediator.GetFileEntry()
    file_system = file_entry.GetFileSystem()

    path_segments = file_system.SplitPath(file_entry.path_spec.location)
    # TODO: validate format of layer identifier.
    return path_segments[-2]

  def _ParseRecord(self, parser_mediator, json_dict):
    """Parses a Docker layer configuration record.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      json_dict (dict): JSON dictionary of the configuration record.
    """
    command = None
    container_configuration = self._GetJSONValue(json_dict, 'container_config')
    if container_configuration:
      command_arguments = self._GetJSONValue(container_configuration, 'Cmd')
      if command_arguments:
        command = ' '.join([
            command_argument.replace('\t', '').strip()
            for command_argument in command_arguments])

    event_data = DockerLayerConfigurationEventData()
    event_data.command = command
    event_data.creation_time = self._ParseISO8601DateTimeString(
        parser_mediator, json_dict, 'created')
    event_data.layer_identifier = self._GetLayerIdentifierFromPath(
        parser_mediator)

    parser_mediator.ProduceEventData(event_data)

  def CheckRequiredFormat(self, json_dict):
    """Check if the record has the minimal structure required by the plugin.

    Args:
      json_dict (dict): JSON dictionary of the configuration record.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    container_configuration = json_dict.get('container_config') or None
    docker_version = json_dict.get('docker_version') or None

    if None in (container_configuration, docker_version):
      return False

    return True


jsonl_parser.JSONLParser.RegisterPlugin(DockerLayerConfigurationJSONLPlugin)



"""Parser for the Sleuthkit (TSK) bodyfile format.

Sleuthkit version 3 format:
MD5|name|inode|mode_as_string|UID|GID|size|atime|mtime|ctime|crtime
0|/lost+found|11|d/drwx------|0|0|12288|1337961350|1337961350|1337961350|0

More information about the format specifications can be read here:
  https://forensics.wiki/bodyfile
"""

import re

from dfdatetime import posix_time as dfdatetime_posix_time

from dfvfs.helpers import text_file

from plaso.containers import events
from plaso.lib import definitions
from plaso.lib import errors
from plaso.parsers import interface
from plaso.parsers import manager


class BodyfileEventData(events.EventData):
  """Bodyfile event data.

  Attributes:
    access_time (dfdatetime.DateTimeValues): file entry last access date
        and time.
    change_time (dfdatetime.DateTimeValues): file entry inode change
        (or metadata last modification) date and time.
    creation_time (dfdatetime.DateTimeValues): file entry creation date
        and time.
    filename (str): name of the file.
    group_identifier (int): group identifier (GID), equivalent to st_gid.
    inode (int): "inode" of the file. Note that inode is an overloaded term
        in the context of a bodyfile and used for MFT entry index values as
        well.
    md5 (str): MD5 hash of the file content, formatted as a hexadecimal string.
    mode_as_string (str): protection mode.
    modification_time (dfdatetime.DateTimeValues): file entry last modification
        date and time.
    offset (int): number of the corresponding line, from which the event data
        was extracted.
    owner_identifier (str): user identifier (UID or SID) of the owner.
    size (int): size of the file content.
    symbolic_link_target (str): path of the symbolic link target.
  """

  DATA_TYPE = 'fs:bodyfile:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.access_time = None
    self.change_time = None
    self.creation_time = None
    self.filename = None
    self.group_identifier = None
    self.inode = None
    self.md5 = None
    self.mode_as_string = None
    self.modification_time = None
    self.offset = None
    self.owner_identifier = None
    self.size = None
    self.symbolic_link_target = None


class BodyfileParser(interface.FileObjectParser):
  """SleuthKit bodyfile parser."""

  NAME = 'bodyfile'
  DATA_FORMAT = 'SleuthKit version 3 bodyfile'

  _INITIAL_FILE_OFFSET = 0

  _UINT32_MAX = (1 << 32) - 1
  _UINT48_MAX = (1 << 48) - 1

  _MD5_RE = re.compile(r'^[0-9a-fA-F]{32}$')

  _NON_PRINTABLE_CHARACTERS = list(range(0, 0x20)) + list(range(0x7f, 0xa0))
  _ESCAPE_CHARACTERS = str.maketrans({
      value: f'\\x{value:02x}' for value in _NON_PRINTABLE_CHARACTERS})

  def _GetDateTimeFromTimestamp(self, float_value):
    """Retrieves a date time object from the floating-point timestamp.

    Args:
      float_value (float): floating-point timestamp in number of seconds since
          January 1, 1970 00:00:00 UTC.

    Returns:
      dfdatetime.TimeElements: date and time based on the floating-point
          timestamp or None if not set.
    """
    if not float_value:
      return None

    integer_value = int(float_value)
    if integer_value == float_value:
      date_time = dfdatetime_posix_time.PosixTime(timestamp=integer_value)
    else:
      integer_value = int(float_value * definitions.NANOSECONDS_PER_SECOND)
      date_time = dfdatetime_posix_time.PosixTimeInNanoseconds(
          timestamp=integer_value)

    date_time.is_local_time = True
    return date_time

  def _GetLastValueAsBase10Integer(
      self, parser_mediator, values, description, line_number, first_line):
    """Retrieves the last value as a base 10 integer.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      values (list[str]): values extracted from the line.
      description (str): human readable description of the value.
      line_number (int): number of the line the values were extracted from.
      first_line (bool): True if this is first line from which values were
          extracted.

    Returns:
      int: integer value or None if not available or invalid.

    Raises:
      WrongParser: when an invalid integer value is found on
          the first line.
    """
    integer_value = values.pop(-1) or None
    if integer_value is not None:
      try:
        integer_value = int(integer_value, 10)
      except ValueError:
        error_string = f'invalid {description:s} value in line: {line_number:d}'
        if first_line:
          raise errors.WrongParser(error_string)

        parser_mediator.ProduceRecoveryWarning(error_string)
        integer_value = None

    return integer_value

  def _GetLastValueAsFloatingPoint(
      self, parser_mediator, values, description, line_number, first_line):
    """Retrieves the last value as floating-point.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      values (list[str]): values extracted from the line.
      description (str): human readable description of the value.
      line_number (int): number of the line the values were extracted from.
      first_line (bool): True if this is first line from which values were
          extracted.

    Returns:
      float: floating-point value or None if not available or invalid.

    Raises:
      WrongParser: when an invalid floating-point value is found on
          the first line.
    """
    float_value = values.pop(-1) or None
    if float_value is not None:
      try:
        float_value = float(float_value)
      except ValueError:
        error_string = f'invalid {description:s} value in line: {line_number:d}'
        if first_line:
          raise errors.WrongParser(error_string)

        parser_mediator.ProduceRecoveryWarning(error_string)
        float_value = None

    return float_value

  def _ParseValues(
      self, parser_mediator, file_offset, line_number, values, first_line):
    """Parses bodyfile values.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_offset (int): offset of the line the values were extracted from,
          relative from the start of the file.
      line_number (int): number of the line the values were extracted from.
      values (list[str]): values extracted from the line.
      first_line (bool): True if this is first line from which values were
          extracted.

    Raises:
      WrongParser: when the values cannot be parsed.
    """
    number_of_values = len(values)
    if number_of_values < 11:
      error_string = (
          f'invalid number of values: {number_of_values:d} in line: '
          f'{line_number:d}')
      if first_line:
        raise errors.WrongParser(error_string)

      parser_mediator.ProduceExtractionWarning(error_string)

      return

    md5_value = values.pop(0)
    if md5_value == '0':
      md5_value = None
    elif md5_value and not self._MD5_RE.match(md5_value):
      error_string = (
          f'invalid MD5 value: {md5_value:s} in line: {line_number:d}')
      if first_line:
        raise errors.WrongParser(error_string)

      parser_mediator.ProduceRecoveryWarning(error_string)

    crtime_value = self._GetLastValueAsFloatingPoint(
        parser_mediator, values, 'creation time', line_number, first_line)
    ctime_value = self._GetLastValueAsFloatingPoint(
        parser_mediator, values, 'inode change time', line_number, first_line)
    mtime_value = self._GetLastValueAsFloatingPoint(
        parser_mediator, values, 'modification time', line_number, first_line)
    atime_value = self._GetLastValueAsFloatingPoint(
        parser_mediator, values, 'access time', line_number, first_line)

    size_value = self._GetLastValueAsBase10Integer(
        parser_mediator, values, 'size', line_number, first_line)
    gid_value = self._GetLastValueAsBase10Integer(
        parser_mediator, values, 'group identifier (GID)', line_number,
        first_line)
    uid_value = self._GetLastValueAsBase10Integer(
        parser_mediator, values, 'user identifier (UID)', line_number,
        first_line)

    if uid_value is not None:
      # Note that the owner_identifier attribute of BodyfileEventData
      # is expected to be a string or None.
      uid_value = f'{uid_value:d}'

    mode_as_string_value = values.pop(-1) or None

    inode_value = values.pop(-1) or None
    if '-' in inode_value:
      inode_value, _, _ = inode_value.partition('-')

    try:
      inode_value = int(inode_value, 10)
    except (TypeError, ValueError):
      inode_value = None
      parser_mediator.ProduceRecoveryWarning(
          f'invalid inode value: {inode_value!s} in line: {line_number:d}')

    # Determine if the inode value is actually a 64-bit NTFS file
    # reference.
    if inode_value > self._UINT48_MAX:
      mft_entry = inode_value & 0xffffffffffff
      if mft_entry <= self._UINT32_MAX:
        inode_value = mft_entry

    filename = '|'.join(values)
    escaped_filename = filename.translate(self._ESCAPE_CHARACTERS)
    if filename != escaped_filename:
      parser_mediator.ProduceRecoveryWarning(
          f'filename in line: {line_number:d} contains unescaped control '
          f'characters')

    else:
      for character in self._NON_PRINTABLE_CHARACTERS:
        escaped_character = f'\\x{character:02x}'
        filename = filename.replace(escaped_character, chr(character))

      filename = filename.replace('\\|', '|')
      filename = filename.replace('\\\\', '\\')

    symbolic_link_target = ''
    if (mode_as_string_value and mode_as_string_value[0] == 'l' and
        ' -> ' in filename):
      filename, _, symbolic_link_target = filename.rpartition(' -> ')

    event_data = BodyfileEventData()
    event_data.access_time = self._GetDateTimeFromTimestamp(atime_value)
    event_data.change_time = self._GetDateTimeFromTimestamp(ctime_value)
    event_data.creation_time = self._GetDateTimeFromTimestamp(crtime_value)
    event_data.filename = filename
    event_data.group_identifier = gid_value
    event_data.inode = inode_value
    event_data.md5 = md5_value
    event_data.mode_as_string = mode_as_string_value
    event_data.modification_time = self._GetDateTimeFromTimestamp(mtime_value)
    event_data.offset = file_offset
    event_data.owner_identifier = uid_value
    event_data.size = size_value
    event_data.symbolic_link_target = symbolic_link_target

    parser_mediator.ProduceEventData(event_data)

  def ParseFileObject(self, parser_mediator, file_object):
    """Parses a bodyfile file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      file_object (dfvfs.FileIO): file-like object.

    Raises:
      WrongParser: when the file cannot be parsed.
    """
    # Note that we cannot use the DSVParser here since the bodyfile format is
    # not strict and clean file format.
    line_reader = text_file.TextFile(
        file_object, encoding='UTF-8', end_of_line='\n')

    first_line = True
    file_offset = 0
    line_number = 0
    number_of_comment_lines = 0

    try:
      line = line_reader.readline()
    except UnicodeDecodeError as exception:
      raise errors.WrongParser(
          f'unable to read line: {line_number:d} with error: {exception!s}')

    while line:
      # Lines that start with '#' are ignored and treated as comments.
      if line[0] == '#':
        number_of_comment_lines += 1

        # It is very uncommon for a bodyfile to have comments, so allow for 10
        # leading comment lines before skipping the file.
        if first_line and number_of_comment_lines > 10:
          raise errors.WrongParser('more than 10 leading comment lines.')

      else:
        values = line.split('|')
        self._ParseValues(
            parser_mediator, file_offset, line_number, values, first_line)

        first_line = False

      file_offset = file_object.tell()
      line_number += 1

      try:
        line = line_reader.readline()
      except UnicodeDecodeError as exception:
        parser_mediator.ProduceExtractionWarning(
            f'unable to read line: {line_number:d} with error: {exception!s}')
        break


manager.ParsersManager.RegisterParser(BodyfileParser)



"""Plist parser plugin for Apple iOS SIM information plist files."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.parsers.plist_plugins import interface
from plaso.parsers.plist import PlistParser


class IOSSIMInfoEventData(events.EventData):
  """Apple iOS SIM information event data.

  Attributes:
    cell_broadcast_version (str): Cell broadcast version.
    eap_aka_enabled (bool): value to indicate EAP-AKA is enabled.
    label_identifier (str): label identifier.
    last_used_time (dfdatetime.DateTimeValues): date and time the SIM was last
        used.
    phone_number (str): phone number.
    sim_type (str): type of SIM
  """

  DATA_TYPE = 'ios:sim:info'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.cell_broadcast_version = None
    self.eap_aka_enabled = None
    self.label_identifier = None
    self.last_used_time = None
    self.phone_number = None
    self.sim_type = None


class IOSSIMInfoPlugin(interface.PlistPlugin):
  """Plist parser plugin for Apple iOS SIM information plist files."""

  NAME = 'ios_siminfo'
  DATA_FORMAT = 'iOS SIM Info plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('com.apple.commcenter.data.plist')])

  # Note that there a com.apple.commcenter.data.plist does not have to contain
  # a PersonalWallet key.
  PLIST_KEYS = frozenset(['PersonalWallet', 'PersonalitySlots', 'pw_ver'])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, match=None, **unused_kwargs):
    """Extract SIM information entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
    """
    personal_wallet = match.get('PersonalWallet', {})
    for _, wallet_item in personal_wallet.items():
      info = wallet_item.get('info', {})
      timestamp = timestamp=info.get('ts')

      event_data = IOSSIMInfoEventData()
      event_data.cell_broadcast_version = info.get('cb_ver')
      event_data.eap_aka_enabled = info.get('eap_aka')
      event_data.label_identifier = info.get('label-id')
      # Mobile Directory Number.
      event_data.phone_number = info.get('mdn')
      event_data.sim_type = info.get('type')

      if timestamp is not None:
        event_data.last_used_time = dfdatetime_posix_time.PosixTime(
            timestamp=timestamp)

      parser_mediator.ProduceEventData(event_data)


PlistParser.RegisterPlugin(IOSSIMInfoPlugin)



"""The language CLI arguments helper."""

from plaso.cli import tools
from plaso.cli.helpers import interface
from plaso.cli.helpers import manager
from plaso.helpers import language_tags
from plaso.lib import errors


class LanguageArgumentsHelper(interface.ArgumentsHelper):
  """Language CLI arguments helper."""

  NAME = 'language'
  DESCRIPTION = 'Language command line arguments.'

  @classmethod
  def AddArguments(cls, argument_group):
    """Adds command line arguments to an argument group.

    This function takes an argument parser or an argument group object and adds
    to it all the command line arguments this helper supports.

    Args:
      argument_group (argparse._ArgumentGroup|argparse.ArgumentParser):
          argparse group.
    """
    argument_group.add_argument(
        '--language', metavar='LANGUAGE_TAG', dest='preferred_language',
        default=None, type=str, help=(
            'The preferred language, which is used for extracting and '
            'formatting Windows EventLog message strings. Use "--language '
            'list" to see a list of supported language tags. The en-US (LCID '
            '0x0409) language is used as fallback if preprocessing could not '
            'determine the system language or no language information is '
            'available in the winevt-rc.db database.'))

  @classmethod
  def ParseOptions(cls, options, configuration_object):
    """Parses and validates options.

    Args:
      options (argparse.Namespace): parser options.
      configuration_object (CLITool): object to be configured by the argument
          helper.

    Raises:
      BadConfigObject: when the configuration object is of the wrong type.
      BadConfigOption: when the language tag is not supported.
    """
    if not isinstance(configuration_object, tools.CLITool):
      raise errors.BadConfigObject(
          'Configuration object is not an instance of CLITool')

    language_tag = cls._ParseStringOption(options, 'preferred_language')

    if (language_tag and language_tag != 'list' and
        not language_tags.LanguageTagHelper.IsLanguageTag(language_tag)):
      raise errors.BadConfigOption(
          f'Unsupported preferred language tag: {language_tag!s}')

    setattr(configuration_object, '_preferred_language', language_tag)


manager.ArgumentHelperManager.RegisterHelper(LanguageArgumentsHelper)



"""Plist parser plugin for Spotlight searched terms plist files.

Fields within the plist key: com.apple.spotlight.plist, where the name of
the key contains the search term.
  DISPLAY_NAME: the display name of the program associated.
  LAST_USED: last time when it was executed.
  PATH: path of the program associated to the term.
"""

from plaso.containers import events
from plaso.parsers import plist
from plaso.parsers.plist_plugins import interface


class SpotlightSearchedTermsEventData(events.EventData):
  """Spotlight searched terms event data.

  Attributes:
    display_name (str): display name.
    last_used_time (dfdatetime.DateTimeValues): last date and time the search
        term was last used.
    path (str): path.
    search_term (str): search term.
  """

  DATA_TYPE = 'spotlight_searched_terms:entry'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.display_name = None
    self.last_used_time = None
    self.path = None
    self.search_term = None


class SpotlightSearchedTermsPlistPlugin(interface.PlistPlugin):
  """Plist parser plugin for Spotlight searched terms plist files."""

  NAME = 'spotlight'
  DATA_FORMAT = 'Spotlight searched terms plist file'

  PLIST_PATH_FILTERS = frozenset([
      interface.PlistPathFilter('com.apple.spotlight.plist')])

  PLIST_KEYS = frozenset(['UserShortcuts'])

  # pylint: disable=arguments-differ
  def _ParsePlist(self, parser_mediator, match=None, **unused_kwargs):
    """Extracts relevant Spotlight entries.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      match (Optional[dict[str: object]]): keys extracted from PLIST_KEYS.
    """
    shortcuts = match.get('UserShortcuts', {})
    for search_term, plist_key in shortcuts.items():
      event_data = SpotlightSearchedTermsEventData()
      event_data.display_name = plist_key.get('DISPLAY_NAME')
      event_data.last_used_time = self._GetDateTimeValueFromPlistKey(
          plist_key, 'LAST_USED')
      event_data.path = plist_key.get('PATH')
      event_data.search_term = search_term

      parser_mediator.ProduceEventData(event_data)


plist.PlistParser.RegisterPlugin(SpotlightSearchedTermsPlistPlugin)



"""Warning attribute containers."""

from acstore.containers import interface
from acstore.containers import manager


class AnalysisWarning(interface.AttributeContainer):
  """Analysis warning attribute container.

  Analysis warnings are produced by analysis plugins when they encounter
  situations that should be brought to the users' attention but are not
  analysis results.

  Attributes:
    message (str): warning message.
    plugin_name (str): name of the analysis plugin to which the warning applies.
  """
  CONTAINER_TYPE = 'analysis_warning'

  SCHEMA = {
      'message': 'str',
      'plugin_name': 'str'}

  def __init__(self, message=None, plugin_name=None):
    """Initializes an analysis warning.

    Args:
      message (Optional[str]): warning message.
      plugin_name (Optional[str]): name of the analysis plugin to which the
          warning applies.
    """
    super().__init__()
    self.message = message
    self.plugin_name = plugin_name


class ExtractionWarning(interface.AttributeContainer):
  """Extraction warning attribute container.

  Extraction warnings are produced by parsers/plugins when they encounter
  situations that should be brought to the users' attention but are not
  event data derived from the data being processed.

  Attributes:
    message (str): warning message.
    parser_chain (str): parser chain to which the warning applies.
    path_spec (dfvfs.PathSpec): path specification of the file entry to which
        the warning applies.
  """
  CONTAINER_TYPE = 'extraction_warning'

  SCHEMA = {
      'message': 'str',
      'parser_chain': 'str',
      'path_spec': 'dfvfs.PathSpec'}

  def __init__(self, message=None, parser_chain=None, path_spec=None):
    """Initializes an extraction warning.

    Args:
      message (Optional[str]): warning message.
      parser_chain (Optional[str]): parser chain to which the warning applies.
      path_spec (Optional[dfvfs.PathSpec]): path specification of the file entry
          to which the warning applies.
    """
    super().__init__()
    self.message = message
    self.parser_chain = parser_chain
    self.path_spec = path_spec


class PreprocessingWarning(interface.AttributeContainer):
  """Preprocessing warning attribute container.

  Preprocessing warnings are produced by preprocessing plugins when they
  encounter situations that should be brought to the users' attention but are
  not preprocessing results.

  Attributes:
    message (str): warning message.
    path_spec (dfvfs.PathSpec): path specification of the file entry to which
        the warning applies.
    plugin_name (str): name of the preprocessing plugin to which the warning
        applies.
  """
  CONTAINER_TYPE = 'preprocessing_warning'

  SCHEMA = {
      'message': 'str',
      'path_spec': 'dfvfs.PathSpec',
      'plugin_name': 'str'}

  def __init__(self, message=None, path_spec=None, plugin_name=None):
    """Initializes an extraction warning.

    Args:
      message (Optional[str]): warning message.
      path_spec (Optional[dfvfs.PathSpec]): path specification of the file entry
          to which the warning applies.
      plugin_name (Optional[str]): name of the preprocessing plugin to which the
          warning applies.
    """
    super().__init__()
    self.message = message
    self.path_spec = path_spec
    self.plugin_name = plugin_name


class RecoveryWarning(interface.AttributeContainer):
  """Recovery warning attribute container.

  Recovery warnings are warning encountered during recovery. They are typically
  produced by parsers/plugins when they are unable to recover event data.

  Attributes:
    message (str): warning message.
    parser_chain (str): parser chain to which the warning applies.
    path_spec (dfvfs.PathSpec): path specification of the file entry to which
        the warning applies.
  """
  CONTAINER_TYPE = 'recovery_warning'

  SCHEMA = {
      'message': 'str',
      'parser_chain': 'str',
      'path_spec': 'dfvfs.PathSpec'}

  def __init__(self, message=None, parser_chain=None, path_spec=None):
    """Initializes a recovery warning.

    Args:
      message (Optional[str]): warning message.
      parser_chain (Optional[str]): parser chain to which the warning applies.
      path_spec (Optional[dfvfs.PathSpec]): path specification of the file entry
          to which the warning applies.
    """
    super().__init__()
    self.message = message
    self.parser_chain = parser_chain
    self.path_spec = path_spec


class TimeliningWarning(interface.AttributeContainer):
  """Timelining warning attribute container.

  Timelining warnings are produced by the timeliner when it encounters
  situations that should be brought to the users' attention but are not
  events derived from the event data being processed.

  Attributes:
    message (str): warning message.
    parser_chain (str): parser chain to which the warning applies.
    path_spec (dfvfs.PathSpec): path specification of the file entry to which
        the warning applies.
  """
  CONTAINER_TYPE = 'timelining_warning'

  SCHEMA = {
      'message': 'str',
      'parser_chain': 'str',
      'path_spec': 'dfvfs.PathSpec'}

  def __init__(self, message=None, parser_chain=None, path_spec=None):
    """Initializes a timelining warning.

    Args:
      message (Optional[str]): warning message.
      parser_chain (Optional[str]): parser chain to which the warning applies.
      path_spec (Optional[dfvfs.PathSpec]): path specification of the file entry
          to which the warning applies.
    """
    super().__init__()
    self.message = message
    self.parser_chain = parser_chain
    self.path_spec = path_spec


manager.AttributeContainersManager.RegisterAttributeContainers([
    AnalysisWarning, ExtractionWarning, PreprocessingWarning, RecoveryWarning,
    TimeliningWarning])



"""The output and formatting multi-processing engine."""

import heapq
import os

from plaso.containers import events
from plaso.engine import processing_status
from plaso.lib import bufferlib
from plaso.lib import definitions
from plaso.lib import errors
from plaso.multi_process import engine
from plaso.multi_process import logger
from plaso.output import mediator as output_mediator
from plaso.storage import time_range as storage_time_range


class PsortEventHeap:
  """Psort event heap."""

  def __init__(self):
    """Initializes a psort events heap."""
    super().__init__()
    self._heap = []

  @property
  def number_of_events(self):
    """int: number of events on the heap."""
    return len(self._heap)

  def PopEvent(self):
    """Pops an event from the heap.

    Returns:
      tuple: containing:

        str: identifier of the event MACB group or None if the event cannot
            be grouped.
        str: identifier of the event content.
        EventObject: event.
        EventData: event data.
        EventDataStream: event data stream.
    """
    try:
      (event_values_hash, _, event, event_data,
       event_data_stream) = heapq.heappop(self._heap)
      return event_values_hash, event, event_data, event_data_stream

    except IndexError:
      return None

  def PopEvents(self):
    """Pops events from the heap.

    Yields:
      tuple: containing:

        str: identifier of the event MACB group or None if the event cannot
            be grouped.
        str: identifier of the event content.
        EventObject: event.
        EventData: event data.
        EventDataStream: event data stream.
    """
    heap_values = self.PopEvent()
    while heap_values:
      yield heap_values
      heap_values = self.PopEvent()

  def PushEvent(self, event, event_data, event_data_stream):
    """Pushes an event onto the heap.

    Args:
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
    """
    event_values_hash = getattr(event_data, '_event_values_hash', None)
    if event_values_hash is None:
      logger.warning('Missing _event_values_hash attribute')
      event_values_hash = 'UNKNOWN'

    timestamp_desc = event.timestamp_desc
    if timestamp_desc is None:
      logger.warning('Missing timestamp_desc attribute')
      timestamp_desc = definitions.TIME_DESCRIPTION_UNKNOWN

    # Note that only events with the same timestamp are stored in the event
    # heap. The event values hash is stored first to cluster events with
    # similar event values.
    heapq.heappush(self._heap, (
        event_values_hash, timestamp_desc, event, event_data,
        event_data_stream))


class OutputAndFormattingMultiProcessEngine(engine.MultiProcessEngine):
  """Output and formatting multi-processing engine."""

  # TODO: move this to a single process engine.
  # pylint: disable=abstract-method

  _HEAP_MAXIMUM_EVENTS = 100000

  _MESSAGE_FORMATTERS_DIRECTORY_NAME = 'formatters'

  _MESSAGE_FORMATTERS_FILE_NAME = 'formatters.yaml'

  def __init__(self):
    """Initializes an output and formatting multi-processing engine."""
    super().__init__()
    # The export event heap is used to make sure the events are sorted in
    # a deterministic way.
    self._events_status = processing_status.EventsStatus()
    self._export_event_heap = PsortEventHeap()
    self._export_event_timestamp = 0
    self._number_of_consumed_events = 0
    self._output_mediator = None
    self._processing_configuration = None
    self._status = definitions.STATUS_INDICATOR_IDLE
    self._status_update_callback = None

  def _CreateOutputMediator(self, storage_reader, processing_configuration):
    """Creates an output mediator.

    Args:
      storage_reader (StorageReader): storage reader.
      processing_configuration (ProcessingConfiguration): processing
          configuration.

    Returns:
      OutputMediator: mediates interactions between output modules and other
          components, such as storage and dfVFS.

    Raises:
      BadConfigOption: if the message formatters file or directory cannot be
          read.
    """
    mediator = output_mediator.OutputMediator(
        storage_reader, data_location=processing_configuration.data_location,
        dynamic_time=processing_configuration.dynamic_time,
        preferred_encoding=processing_configuration.preferred_encoding)

    if processing_configuration.preferred_language:
      try:
        mediator.SetPreferredLanguageIdentifier(
            processing_configuration.preferred_language)
      except (KeyError, TypeError):
        logger.warning((
            f'Unable to to set preferred language: '
            f'{processing_configuration.preferred_language!s}'))

    mediator.SetTimeZone(processing_configuration.preferred_time_zone)

    self._ReadMessageFormatters(
        mediator, processing_configuration.data_location,
        processing_configuration.custom_formatters_path)

    return mediator

  def _ExportEvent(
      self, storage_reader, output_module, event, event_data, event_data_stream,
      deduplicate_events=True):
    """Exports an event using an output module.

    Args:
      storage_reader (StorageReader): storage reader.
      output_module (OutputModule): output module.
      event (EventObject): event.
      event_data (EventData): event data.
      event_data_stream (EventDataStream): event data stream.
      deduplicate_events (Optional[bool]): True if events should be
          deduplicated.
    """
    if (event.timestamp != self._export_event_timestamp or
        self._export_event_heap.number_of_events > self._HEAP_MAXIMUM_EVENTS):
      self._FlushExportBuffer(
          storage_reader, output_module, deduplicate_events=deduplicate_events)
      self._export_event_timestamp = event.timestamp

    self._export_event_heap.PushEvent(event, event_data, event_data_stream)

  def _ExportEvents(
      self, storage_reader, output_module, deduplicate_events=True,
      event_filter=None, time_slice=None, use_time_slicer=False):
    """Exports events using an output module.

    Args:
      storage_reader (StorageReader): storage reader.
      output_module (OutputModule): output module.
      deduplicate_events (Optional[bool]): True if events should be
          deduplicated.
      event_filter (Optional[EventObjectFilter]): event filter.
      time_slice (Optional[TimeRange]): time range that defines a time slice
          to filter events.
      use_time_slicer (Optional[bool]): True if the 'time slicer' should be
          used. The 'time slicer' will provide a context of events around
          an event of interest.
    """
    self._status = definitions.STATUS_INDICATOR_EXPORTING

    time_slice_buffer = None
    time_slice_range = None

    if time_slice:
      if time_slice.event_timestamp is not None:
        time_slice_range = storage_time_range.TimeRange(
            time_slice.start_timestamp, time_slice.end_timestamp)

      if use_time_slicer:
        time_slice_buffer = bufferlib.CircularBuffer(time_slice.duration)

    filter_limit = getattr(event_filter, 'limit', None)
    forward_entries = 0

    self._events_status.number_of_filtered_events = 0
    self._events_status.number_of_events_from_time_slice = 0

    for event in storage_reader.GetSortedEvents(time_range=time_slice_range):
      event_data_identifier = event.GetEventDataIdentifier()
      event_data = storage_reader.GetAttributeContainerByIdentifier(
          events.EventData.CONTAINER_TYPE, event_data_identifier)

      event_data_stream_identifier = event_data.GetEventDataStreamIdentifier()
      if event_data_stream_identifier:
        event_data_stream = storage_reader.GetAttributeContainerByIdentifier(
            events.EventDataStream.CONTAINER_TYPE, event_data_stream_identifier)
      else:
        event_data_stream = None

      event_identifier = event.GetIdentifier()
      event_tag = storage_reader.GetEventTagByEventIdentifer(event_identifier)

      if time_slice_range and event.timestamp != time_slice.event_timestamp:
        self._events_status.number_of_events_from_time_slice += 1

      if event_filter:
        filter_match = event_filter.Match(
            event, event_data, event_data_stream, event_tag)
      else:
        filter_match = None

      # pylint: disable=singleton-comparison
      if filter_match == False:
        if not time_slice_buffer:
          self._events_status.number_of_filtered_events += 1

        elif forward_entries == 0:
          time_slice_buffer.Append((event, event_data))
          self._events_status.number_of_filtered_events += 1

        elif forward_entries <= time_slice_buffer.size:
          self._ExportEvent(
              storage_reader, output_module, event, event_data,
              event_data_stream, deduplicate_events=deduplicate_events)
          self._number_of_consumed_events += 1
          self._events_status.number_of_events_from_time_slice += 1
          forward_entries += 1

        else:
          # We reached the maximum size of the time slice and don't need to
          # include other entries.
          self._events_status.number_of_filtered_events += 1
          forward_entries = 0

      else:
        # pylint: disable=singleton-comparison
        if filter_match == True and time_slice_buffer:
          # Empty the time slice buffer.
          for event_in_buffer, event_data_in_buffer in (
              time_slice_buffer.Flush()):
            self._ExportEvent(
                storage_reader, output_module, event_in_buffer,
                event_data_in_buffer, event_data_stream,
                deduplicate_events=deduplicate_events)
            self._number_of_consumed_events += 1
            self._events_status.number_of_filtered_events += 1
            self._events_status.number_of_events_from_time_slice += 1

          forward_entries = 1

        self._ExportEvent(
            storage_reader, output_module, event, event_data, event_data_stream,
            deduplicate_events=deduplicate_events)
        self._number_of_consumed_events += 1

        # pylint: disable=singleton-comparison
        if (filter_match == True and filter_limit and
            filter_limit == self._number_of_consumed_events):
          break

    self._FlushExportBuffer(storage_reader, output_module)

  def _FlushExportBuffer(
      self, storage_reader, output_module, deduplicate_events=True):
    """Flushes buffered events and writes them to the output module.

    Args:
      storage_reader (StorageReader): storage reader.
      output_module (OutputModule): output module.
      deduplicate_events (Optional[bool]): True if events should be
          deduplicated.
    """
    last_event_values_hash = None
    last_macb_group_identifier = None
    last_timestamp_desc = None
    macb_group = []

    for (event_values_hash, event, event_data,
         event_data_stream) in self._export_event_heap.PopEvents():
      timestamp_desc = event.timestamp_desc

      if (deduplicate_events and timestamp_desc == last_timestamp_desc and
          event_values_hash == last_event_values_hash):
        self._events_status.number_of_duplicate_events += 1
        continue

      event_identifier = event.GetIdentifier()
      event_tag = storage_reader.GetEventTagByEventIdentifer(event_identifier)

      if timestamp_desc in (
          definitions.TIME_DESCRIPTION_LAST_ACCESS,
          definitions.TIME_DESCRIPTION_CREATION,
          definitions.TIME_DESCRIPTION_METADATA_MODIFICATION,
          definitions.TIME_DESCRIPTION_MODIFICATION):
        macb_group_identifier = event_values_hash
      else:
        macb_group_identifier = None

      if macb_group_identifier is None:
        if macb_group:
          output_module.WriteFieldValuesOfMACBGroup(
              self._output_mediator, macb_group)
          macb_group = []

        field_values = output_module.GetFieldValues(
            self._output_mediator, event, event_data, event_data_stream,
            event_tag)
        output_module.WriteFieldValues(self._output_mediator, field_values)

      else:
        if (last_macb_group_identifier == macb_group_identifier or
            not macb_group):
          macb_group.append((event, event_data, event_data_stream, event_tag))

        else:
          output_module.WriteFieldValuesOfMACBGroup(
              self._output_mediator, macb_group)
          macb_group = [(event, event_data, event_data_stream, event_tag)]

        self._events_status.number_of_macb_grouped_events += 1

      last_event_values_hash = event_values_hash
      last_macb_group_identifier = macb_group_identifier
      last_timestamp_desc = timestamp_desc

    if macb_group:
      output_module.WriteFieldValuesOfMACBGroup(
          self._output_mediator, macb_group)

  def _ReadMessageFormatters(
      self, output_mediator_object, data_location, custom_formatters_path):
    """Reads the message formatters from a formatters file or directory.

    Args:
      output_mediator_object (OutputMediator): mediates interactions between
          output modules and other components, such as storage and dfVFS.
      data_location (str): path to the data files.
      custom_formatters_path (str): path to custom formatter definitions file.

    Raises:
      BadConfigOption: if the message formatters file or directory cannot be
          read.
    """
    formatters_directory = os.path.join(
        data_location, self._MESSAGE_FORMATTERS_DIRECTORY_NAME)
    formatters_file = os.path.join(
        data_location, self._MESSAGE_FORMATTERS_FILE_NAME)

    if os.path.isdir(formatters_directory):
      try:
        output_mediator_object.ReadMessageFormattersFromDirectory(
            formatters_directory)
      except KeyError as exception:
        raise errors.BadConfigOption((
            f'Unable to read message formatters from directory: '
            f'{formatters_directory:s} with error: {exception!s}'))

    elif os.path.isfile(formatters_file):
      try:
        output_mediator_object.ReadMessageFormattersFromFile(formatters_file)
      except KeyError as exception:
        raise errors.BadConfigOption((
            f'Unable to read message formatters from file: '
            f'{formatters_file:s} with error: {exception!s}'))

    else:
      raise errors.BadConfigOption('Missing formatters file and directory.')

    if custom_formatters_path:
      try:
        output_mediator_object.ReadMessageFormattersFromFile(
            custom_formatters_path, override_existing=True)
      except KeyError as exception:
        raise errors.BadConfigOption((
            f'Unable to read custrom message formatters from file: '
            f'{formatters_file:s} with error: {exception!s}'))

  def _UpdateForemanProcessStatus(self):
    """Update the foreman process status."""
    used_memory = self._process_information.GetUsedMemory() or 0

    self._processing_status.UpdateForemanStatus(
        self._name, self._status, self._pid, used_memory, '',
        0, 0, 0, 0, self._number_of_consumed_events, 0, 0, 0, 0, 0)

    self._processing_status.UpdateEventsStatus(self._events_status)

  def _UpdateStatus(self):
    """Update the status."""
    self._UpdateForemanProcessStatus()

    if self._status_update_callback:
      self._status_update_callback(self._processing_status)

  def ExportEvents(
      self, storage_reader, output_module, processing_configuration,
      deduplicate_events=True, event_filter=None, status_update_callback=None,
      time_slice=None, use_time_slicer=False):
    """Exports events using an output module.

    Args:
      storage_reader (StorageReader): storage reader.
      output_module (OutputModule): output module.
      processing_configuration (ProcessingConfiguration): processing
          configuration.
      deduplicate_events (Optional[bool]): True if events should be
          deduplicated.
      event_filter (Optional[EventObjectFilter]): event filter.
      status_update_callback (Optional[function]): callback function for status
          updates.
      time_slice (Optional[TimeSlice]): slice of time to output.
      use_time_slicer (Optional[bool]): True if the 'time slicer' should be
          used. The 'time slicer' will provide a context of events around
          an event of interest.

    Raises:
      BadConfigOption: if the message formatters file or directory cannot be
          read.
    """
    self._events_status = processing_status.EventsStatus()
    self._processing_configuration = processing_configuration
    self._status_update_callback = status_update_callback

    total_number_of_events = 0
    if storage_reader.HasAttributeContainers('parser_count'):
      parsers_counter = {
          parser_count.name: parser_count.number_of_events
          for parser_count in storage_reader.GetAttributeContainers(
              'parser_count')}

      total_number_of_events = parsers_counter['total']

    self._events_status.total_number_of_events = total_number_of_events

    self._output_mediator = self._CreateOutputMediator(
        storage_reader, processing_configuration)

    output_module.WriteHeader(output_mediator)

    self._StartStatusUpdateThread()

    self._StartProfiling(self._processing_configuration.profiling)

    try:
      self._ExportEvents(
          storage_reader, output_module, deduplicate_events=deduplicate_events,
          event_filter=event_filter, time_slice=time_slice,
          use_time_slicer=use_time_slicer)

      self._status = definitions.STATUS_INDICATOR_COMPLETED

    finally:
      # Stop the status update thread after close of the storage writer
      # so we include the storage sync to disk in the status updates.
      self._StopStatusUpdateThread()

    output_module.WriteFooter()

    # Update the status view one last time.
    self._UpdateStatus()

    self._StopProfiling()

    # Reset values.
    self._events_status = None
    self._output_mediator = None
    self._processing_configuration = None
    self._status_update_callback = None



"""The OpenSearch Timesketch output module CLI arguments helper."""

from plaso.cli.helpers import interface
from plaso.cli.helpers import manager
from plaso.cli.helpers import opensearch_output
from plaso.lib import errors
from plaso.output import opensearch_ts


class OpenSearchTimesketchOutputArgumentsHelper(interface.ArgumentsHelper):
  """OpenSearch Timesketch output module CLI arguments helper."""

  NAME = 'opensearch_ts'
  CATEGORY = 'output'
  DESCRIPTION = 'Argument helper for the OpenSearch Timesketch output module.'

  _DEFAULT_TIMELINE_IDENTIFIER = 0

  @classmethod
  def AddArguments(cls, argument_group):
    """Adds command line arguments the helper supports to an argument group.

    This function takes an argument parser or an argument group object and adds
    to it all the command line arguments this helper supports.

    Args:
      argument_group (argparse._ArgumentGroup|argparse.ArgumentParser):
          argparse group.
    """
    opensearch_output.OpenSearchOutputArgumentsHelper.AddArguments(
        argument_group)

    argument_group.add_argument(
        '--timeline_identifier', '--timeline-identifier',
        dest='timeline_identifier', type=int,
        default=cls._DEFAULT_TIMELINE_IDENTIFIER, action='store',
        metavar='IDENTIFIER', help=(
             'The identifier of the timeline in Timesketch.'))

  @classmethod
  def ParseOptions(cls, options, output_module):  # pylint: disable=arguments-renamed
    """Parses and validates options.

    Args:
      options (argparse.Namespace): parser options.
      output_module (OutputModule): output module to configure.

    Raises:
      BadConfigObject: when the output module object is of the wrong type.
      BadConfigOption: when a configuration parameter fails validation.
    """
    if not isinstance(
        output_module, opensearch_ts.OpenSearchTimesketchOutputModule):
      raise errors.BadConfigObject(
          'Output module is not an instance of OpenSearchsearchOutputModule')

    opensearch_output.OpenSearchOutputArgumentsHelper.ParseOptions(
        options, output_module)

    timeline_identifier = cls._ParseNumericOption(
        options, 'timeline_identifier',
        default_value=cls._DEFAULT_TIMELINE_IDENTIFIER)

    if timeline_identifier:
      output_module.SetTimelineIdentifier(timeline_identifier)


manager.ArgumentHelperManager.RegisterHelper(
    OpenSearchTimesketchOutputArgumentsHelper)



"""Imports for the plist parser plugins."""

from plaso.parsers.plist_plugins import airport
from plaso.parsers.plist_plugins import apple_account
from plaso.parsers.plist_plugins import bluetooth
from plaso.parsers.plist_plugins import default
from plaso.parsers.plist_plugins import install_history
from plaso.parsers.plist_plugins import ios_carplay
from plaso.parsers.plist_plugins import ios_identityservices
from plaso.parsers.plist_plugins import ios_mobilebackup
from plaso.parsers.plist_plugins import ios_siminfo
from plaso.parsers.plist_plugins import ios_wifi_known_networks
from plaso.parsers.plist_plugins import ipod
from plaso.parsers.plist_plugins import launchd
from plaso.parsers.plist_plugins import macos_background_items
from plaso.parsers.plist_plugins import macos_login_items
from plaso.parsers.plist_plugins import macos_login_window
from plaso.parsers.plist_plugins import macos_startup_item
from plaso.parsers.plist_plugins import macos_user
from plaso.parsers.plist_plugins import safari_downloads
from plaso.parsers.plist_plugins import safari_history
from plaso.parsers.plist_plugins import software_update
from plaso.parsers.plist_plugins import spotlight_searched_terms
from plaso.parsers.plist_plugins import spotlight_volume
from plaso.parsers.plist_plugins import time_machine



"""Text file parser plugin for Santa log files."""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements
from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class SantaMountEventData(events.EventData):
  """Santa mount event data.

  Attributes:
    action (str): event type recorded by Santa.
    appearance_time (dfdatetime.DateTimeValues): date and time the disk
        appeared.
    bsd_name (str): disk BSD name.
    bus (str): device protocol.
    dmg_path (str): DMG file path.
    fs (str): disk volume kind.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    model (str): disk model.
    mount (str): disk mount point.
    serial (str): disk serial.
    volume (str): disk volume name.
  """

  DATA_TYPE = 'santa:diskmount'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.action = None
    self.appearance_time = None
    self.bsd_name = None
    self.bus = None
    self.dmg_path = None
    self.fs = None
    self.last_written_time = None
    self.model = None
    self.mount = None
    self.serial = None
    self.volume = None


class SantaExecutionEventData(events.EventData):
  """Santa execution event data.

  Attributes:
    action (str): action recorded by Santa.
    certificate_common_name (str): certificate common name.
    certificate_hash (str): SHA256 hash for the certificate associated with the
        executed process.
    decision (str): if the process was allowed or blocked.
    gid (str): group identifier associated with the executed process.
    group (str): group name associated with the executed process.
    last_run_time (dfdatetime.DateTimeValues): executable (binary) last run
        date and time.
    long_reason (str): further explanation behind Santa decision to execute
        or block a process.
    mode (str): Santa execution mode, for example Monitor or Lockdown.
    pid (str): process identifier for the process.
    pid_version (str): the process identifier version extracted from the Mach
        audit token. The version can sed to identify process identifier
        rollovers.
    ppid (str): parent process identifier for the executed process.
    process_arguments (str): executed process with its arguments.
    process_hash (str): SHA256 hash for the executed process.
    process_path (str): process file path.
    reason (str): reason behind Santa decision to execute or block a process.
    uid (str): user identifier associated with the executed process.
    user (str): user name associated with the executed process.
  """

  DATA_TYPE = 'santa:execution'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.action = None
    self.certificate_common_name = None
    self.certificate_hash = None
    self.decision = None
    self.gid = None
    self.group = None
    self.last_run_time = None
    self.long_reason = None
    self.mode = None
    self.pid = None
    self.pid_version = None
    self.ppid = None
    self.process_arguments = None
    self.process_hash = None
    self.process_path = None
    self.quarantine_url = None
    self.reason = None
    self.uid = None
    self.user = None


class SantaFileSystemEventData(events.EventData):
  """Santa file system event data.

  Attributes:
    action (str): event type recorded by Santa.
    file_new_path (str): new file path and name for RENAME events.
    file_path (str): file path and name for WRITE/DELETE events.
    gid (str): group identifier associated with the executed process.
    group (str): group name associated with the executed process.
    last_written_time (dfdatetime.DateTimeValues): entry last written date and
        time.
    pid (str): process identifier for the process.
    pid_version (str): the process identifier version extracted from the Mach
        audit token. The version can be used to identify process identifier
        rollovers.
    ppid (str): parent process identifier for the executed process.
    process_path (str): process file path.
    process (str): process name.
    uid (str): user identifier associated with the executed process.
    user (str): user name associated with the executed process.
  """

  DATA_TYPE = 'santa:file_system_event'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.action = None
    self.file_new_path = None
    self.file_path = None
    self.gid = None
    self.group = None
    self.last_written_time = None
    self.pid = None
    self.pid_version = None
    self.ppid = None
    self.process = None
    self.process_path = None
    self.uid = None
    self.user = None


class SantaProcessExitEventData(events.EventData):
  """Santa process exit event data.

  Attributes:
    action (str): action recorded by Santa.
    exit_time (dfdatetime.DateTimeValues): process exit date and time.
    gid (str): group identifier associated with the executed process.
    pid (str): process identifier for the process.
    pid_version (str): the process identifier version extracted from the Mach
        audit token. The version can be used to identify process identifier
        rollovers.
    ppid (str): parent process identifier for the executed process.
    uid (str): user identifier associated with the executed process.
  """

  DATA_TYPE = 'santa:process_exit'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.action = None
    self.exit_time = None
    self.gid = None
    self.pid = None
    self.pid_version = None
    self.ppid = None
    self.uid = None


class SantaTextPlugin(interface.TextPlugin):
  """Text file parser plugin for Santa log files."""

  NAME = 'santa'
  DATA_FORMAT = 'Santa log (santa.log) file'

  ENCODING = 'utf-8'

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_DIGITS = pyparsing.Word(pyparsing.nums, exact=3).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _SEPARATOR = pyparsing.Suppress('|')

  # Using CharsNotIn is more efficient than SkipTo('|'). Optional is needed
  # since CharsNotIn does not support a minimum of 0.
  _SKIP_TO_SEPARATOR = pyparsing.Optional(pyparsing.CharsNotIn('|\n'))

  # Date and time values are formatted as: 2018-08-19T03:09:13.120Z
  _DATE_AND_TIME = (
      _FOUR_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('T') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress(':') +
      _TWO_DIGITS + pyparsing.Suppress('.') +
      _THREE_DIGITS + pyparsing.Suppress('Z')).set_results_name('date_time')

  _DATE_TIME_BLOCK = (
      pyparsing.Suppress('[') + _DATE_AND_TIME + pyparsing.Suppress(']'))

  _END_OF_LINE = pyparsing.Suppress(pyparsing.LineEnd())

  _SANTAD_PREAMBLE = pyparsing.Suppress('I santad:')

  _APPEARANCE = (pyparsing.Suppress('|appearance=') +
                 _DATE_AND_TIME.set_results_name('appearance'))

  _ARGS = (pyparsing.Suppress('|args=') +
           _SKIP_TO_SEPARATOR.set_results_name('args'))

  _BSD_NAME = (pyparsing.Suppress('|bsdname=') +
               _SKIP_TO_SEPARATOR.set_results_name('bsd_name'))

  _BUS = pyparsing.Suppress('|bus=') + _SKIP_TO_SEPARATOR.set_results_name(
      'bus')

  _CERT_CN = (pyparsing.Suppress('|cert_cn=') +
              _SKIP_TO_SEPARATOR.set_results_name('cert_cn'))

  _CERT_SHA256 = (pyparsing.Suppress('|cert_sha256=') +
                  _SKIP_TO_SEPARATOR.set_results_name('cert_sha256'))

  _DECISION = (pyparsing.Suppress('|decision=') +
               _SKIP_TO_SEPARATOR.set_results_name('decision'))

  _DMG_PATH = (pyparsing.Suppress('|dmgpath=') +
               _SKIP_TO_SEPARATOR.set_results_name('dmg_path'))

  _EXPLAIN = (pyparsing.Suppress('|explain=') +
              _SKIP_TO_SEPARATOR.set_results_name('explain'))

  _FILE_SYSTEM = (pyparsing.Suppress('|fs=') +
                  _SKIP_TO_SEPARATOR.set_results_name('fs'))

  _GID = (pyparsing.Suppress('|gid=') +
          _SKIP_TO_SEPARATOR.set_results_name('gid'))

  _GROUP = (pyparsing.Suppress('|group=') +
            _SKIP_TO_SEPARATOR.set_results_name('group'))

  _MODE = (pyparsing.Suppress('|mode=') +
           _SKIP_TO_SEPARATOR.set_results_name('mode'))

  _MODEL = (pyparsing.Suppress('|model=') +
            _SKIP_TO_SEPARATOR.set_results_name('model'))

  _MOUNT = (pyparsing.Suppress('|mount=') +
            _SKIP_TO_SEPARATOR.set_results_name('mount'))

  _NEW_PATH = (pyparsing.Suppress('|newpath=') +
               _SKIP_TO_SEPARATOR.set_results_name('newpath'))

  _QUARANTINE_URL = (pyparsing.Suppress('|quarantine_url=') +
                     _SKIP_TO_SEPARATOR.set_results_name('quarantine_url'))

  _PATH = (pyparsing.Suppress('|path=') +
           _SKIP_TO_SEPARATOR.set_results_name('path'))

  _PID = (pyparsing.Suppress('|pid=') +
          _SKIP_TO_SEPARATOR.set_results_name('pid'))

  _PID_VERSION = (pyparsing.Suppress('|pidversion=') +
                 _SKIP_TO_SEPARATOR.set_results_name('pidversion'))

  _PPID = (pyparsing.Suppress('|ppid=') +
           _SKIP_TO_SEPARATOR.set_results_name('ppid'))

  _PROCESS = (pyparsing.Suppress('|process=') +
              _SKIP_TO_SEPARATOR.set_results_name('process'))

  _PROCESS_PATH = (pyparsing.Suppress('|processpath=') +
                   _SKIP_TO_SEPARATOR.set_results_name('processpath'))

  _REASON = (pyparsing.Suppress('|reason=') +
             _SKIP_TO_SEPARATOR.set_results_name('reason'))

  _SERIAL = (pyparsing.Suppress('|serial=') +
             _SKIP_TO_SEPARATOR.set_results_name('serial'))

  _SHA256 = (pyparsing.Suppress('|sha256=') +
             _SKIP_TO_SEPARATOR.set_results_name('sha256'))

  _UID = (pyparsing.Suppress('|uid=') +
          _SKIP_TO_SEPARATOR.set_results_name('uid'))

  _USER = (pyparsing.Suppress('|user=') +
           _SKIP_TO_SEPARATOR.set_results_name('user'))

  _VOLUME = (pyparsing.Suppress('|volume=') +
             _SKIP_TO_SEPARATOR.set_results_name('volume'))

  _DISK_MOUNT_LINE = (
      _DATE_TIME_BLOCK + _SANTAD_PREAMBLE + pyparsing.Suppress('action=') +
      pyparsing.Literal('DISKAPPEAR').set_results_name('action') +
      _MOUNT + _VOLUME + _BSD_NAME + _FILE_SYSTEM + _MODEL + _SERIAL + _BUS +
      _DMG_PATH + _APPEARANCE + _END_OF_LINE)

  _DISK_UNMOUNT_LINE = (
      _DATE_TIME_BLOCK + _SANTAD_PREAMBLE + pyparsing.Suppress('action=') +
      pyparsing.Literal('DISKDISAPPEAR').set_results_name('action') +
      _MOUNT + _VOLUME + _BSD_NAME + _END_OF_LINE)

  _EXECUTION_LINE = (
      _DATE_TIME_BLOCK + _SANTAD_PREAMBLE + pyparsing.Suppress('action=') +
      pyparsing.Literal('EXEC').set_results_name('action') +
      _DECISION + _REASON + pyparsing.Optional(_EXPLAIN) + _SHA256 +
      pyparsing.Optional(_CERT_SHA256) + pyparsing.Optional(_CERT_CN) +
      pyparsing.Optional(_QUARANTINE_URL) + _PID +
      pyparsing.Optional(_PID_VERSION) + _PPID + _UID + _USER + _GID + _GROUP +
      _MODE + _PATH + pyparsing.Optional(_ARGS) + _END_OF_LINE)

  _FILE_OPERATION_LINE = (
      _DATE_TIME_BLOCK + _SANTAD_PREAMBLE + pyparsing.Suppress('action=') + (
          pyparsing.Literal('DELETE') ^
          pyparsing.Literal('LINK') ^
          pyparsing.Literal('RENAME') ^
          pyparsing.Literal('WRITE')).set_results_name('action') +
      _PATH + pyparsing.Optional(_NEW_PATH) + _PID +
      pyparsing.Optional(_PID_VERSION) + _PPID + _PROCESS + _PROCESS_PATH +
      _UID + _USER + _GID + _GROUP + _END_OF_LINE)

  _PROCESS_EXIT_LINE = (
      _DATE_TIME_BLOCK + _SANTAD_PREAMBLE + pyparsing.Suppress('action=') +
      pyparsing.Literal('EXIT').set_results_name('action') +
      _PID + _PID_VERSION + _PPID + _UID + _GID + _END_OF_LINE)

  _QUOTA_EXCEEDED_LINE = (
      _DATE_TIME_BLOCK + pyparsing.Literal((
          '*** LOG MESSAGE QUOTA EXCEEDED - SOME MESSAGES FROM THIS PROCESS '
          'HAVE BEEN DISCARDED ***')) + _END_OF_LINE)

  _LINE_STRUCTURES = [
      ('execution_line', _EXECUTION_LINE),
      ('file_system_event_line', _FILE_OPERATION_LINE),
      ('mount_line', _DISK_MOUNT_LINE),
      ('process_exit_line', _PROCESS_EXIT_LINE),
      ('quota_exceeded_line', _QUOTA_EXCEEDED_LINE),
      ('unmount_line', _DISK_UNMOUNT_LINE)]

  VERIFICATION_GRAMMAR = (
      _DISK_MOUNT_LINE ^ _DISK_UNMOUNT_LINE ^ _EXECUTION_LINE ^
      _FILE_OPERATION_LINE ^ _PROCESS_EXIT_LINE)

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key == 'quota_exceeded_line':
      # skip this line
      return

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    date_time = self._ParseTimeElements(time_elements_structure)

    if key == 'execution_line':
      event_data = SantaExecutionEventData()
      event_data.action = self._GetValueFromStructure(structure, 'action')
      event_data.certificate_common_name = self._GetValueFromStructure(
          structure, 'cert_cn')
      event_data.certificate_hash = self._GetValueFromStructure(
          structure, 'cert_sha256')
      event_data.decision = self._GetValueFromStructure(structure, 'decision')
      event_data.gid = self._GetValueFromStructure(structure, 'gid')
      event_data.group = self._GetValueFromStructure(structure, 'group')
      event_data.last_run_time = date_time
      event_data.long_reason = self._GetValueFromStructure(structure, 'explain')
      event_data.mode = self._GetValueFromStructure(structure, 'mode')
      event_data.quarantine_url = self._GetValueFromStructure(
          structure, 'quarantine_url')
      event_data.pid = self._GetValueFromStructure(structure, 'pid')
      event_data.pid_version = self._GetValueFromStructure(
          structure, 'pidversion')
      event_data.ppid = self._GetValueFromStructure(structure, 'ppid')
      event_data.process_arguments = self._GetValueFromStructure(
          structure, 'args')
      event_data.process_hash = self._GetValueFromStructure(structure, 'sha256')
      event_data.process_path = self._GetValueFromStructure(structure, 'path')
      event_data.reason = self._GetValueFromStructure(structure, 'reason')
      event_data.uid = self._GetValueFromStructure(structure, 'uid')
      event_data.user = self._GetValueFromStructure(structure, 'user')

    if key == 'process_exit_line':
      event_data = SantaProcessExitEventData()
      event_data.action = self._GetValueFromStructure(structure, 'action')
      event_data.exit_time = date_time
      event_data.pid = self._GetValueFromStructure(structure, 'pid')
      event_data.pid_version = self._GetValueFromStructure(
          structure, 'pidversion')
      event_data.ppid = self._GetValueFromStructure(structure, 'ppid')
      event_data.uid = self._GetValueFromStructure(structure, 'uid')
      event_data.gid = self._GetValueFromStructure(structure, 'gid')

    elif key == 'file_system_event_line':
      event_data = SantaFileSystemEventData()
      event_data.action = self._GetValueFromStructure(structure, 'action')
      event_data.file_new_path = self._GetValueFromStructure(
          structure, 'newpath')
      event_data.file_path = self._GetValueFromStructure(structure, 'path')
      event_data.gid = self._GetValueFromStructure(structure, 'gid')
      event_data.group = self._GetValueFromStructure(structure, 'group')
      event_data.last_written_time = date_time
      event_data.pid = self._GetValueFromStructure(structure, 'pid')
      event_data.pid_version = self._GetValueFromStructure(
          structure, 'pidversion')
      event_data.ppid = self._GetValueFromStructure(structure, 'ppid')
      event_data.process = self._GetValueFromStructure(structure, 'process')
      event_data.process_path = self._GetValueFromStructure(
          structure, 'processpath')
      event_data.uid = self._GetValueFromStructure(structure, 'uid')
      event_data.user = self._GetValueFromStructure(structure, 'user')

    elif key == 'unmount_line':
      event_data = SantaMountEventData()
      event_data.action = self._GetValueFromStructure(structure, 'action')
      event_data.bsd_name = self._GetValueFromStructure(structure, 'bsd_name')
      event_data.last_written_time = date_time
      event_data.mount = self._GetValueFromStructure(structure, 'mount') or None
      event_data.volume = self._GetValueFromStructure(structure, 'volume')

    elif key == 'mount_line':
      event_data = SantaMountEventData()
      event_data.action = self._GetValueFromStructure(structure, 'action')
      event_data.bsd_name = self._GetValueFromStructure(structure, 'bsd_name')
      event_data.bus = self._GetValueFromStructure(structure, 'bus')
      event_data.dmg_path = self._GetValueFromStructure(structure, 'dmg_path')
      event_data.fs = self._GetValueFromStructure(structure, 'fs')
      event_data.last_written_time = date_time
      event_data.model = self._GetValueFromStructure(structure, 'model')
      event_data.mount = self._GetValueFromStructure(structure, 'mount') or None
      event_data.serial = self._GetValueFromStructure(
          structure, 'serial') or None
      event_data.volume = self._GetValueFromStructure(structure, 'volume')

      appearance = self._GetValueFromStructure(structure, 'appearance')
      if appearance:
        # pylint: disable=attribute-defined-outside-init
        event_data.appearance_time = self._ParseTimeElements(appearance)

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements
          of a log line.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived from
          the time elements.
    """
    try:
      (year, month, day_of_month, hours, minutes, seconds, milliseconds) = (
          time_elements_structure)

      # Ensure time_elements_tuple is not a pyparsing.ParseResults otherwise
      # copy.deepcopy() of the dfDateTime object will fail on Python 3.8 with:
      # "TypeError: 'str' object is not callable" due to pyparsing.ParseResults
      # overriding __getattr__ with a function that returns an empty string
      # when named token does not exist.
      time_elements_tuple = (
          year, month, day_of_month, hours, minutes, seconds, milliseconds)

      return dfdatetime_time_elements.TimeElementsInMilliseconds(
          time_elements_tuple=time_elements_tuple)

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(SantaTextPlugin)



"""Text parser plugin for Atlassian Bitbucket access log files.

This is for the atlassian-bitbucket-access.log file, one of multiple log
files produced by a Bitbucket DC/Server installation.

The standard HTTP/SSH access log format is pipe-delimited:
  ip_address | protocol | request_identifier | user | timestamp | "request" |
  "referer" "user-agent" | status | bytes_read | bytes_written | labels |
  request_time | session_identifier |

The Mesh/gRPC access log format has two additional fields:
  ip_address | grpc | request_identifier | mesh_execution_identifier | user |
  timestamp | "action" | - | status | bytes_read | bytes_written | mesh_in |
  mesh_out | duration_ns | labels | session_identifier |

Also see:
  https://support.atlassian.com/bitbucket-data-center/kb/how-to-read-the-bitbucket-data-center-log-formats/
"""

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.lib import errors
from plaso.parsers import text_parser
from plaso.parsers.text_plugins import interface


class BitbucketAccessEventData(events.EventData):
  """Bitbucket access log event data.

  Attributes:
    http_request_method (str): HTTP request method (GET, POST, etc.), or
        'SSH' for SSH requests, or None for gRPC/Mesh action lines.
    http_request_uri (str): HTTP request URI, SSH command, or gRPC action.
    http_request_user_agent (str): HTTP request user agent, or None for
        SSH/gRPC requests.
    http_response_bytes_read (int): number of bytes read from the client.
    http_response_bytes_written (int): number of bytes written to the client.
    http_response_code (int): HTTP response status code.
    http_version (str): HTTP request version, or None for SSH/gRPC requests.
    labels (str): request classification labels, such as push, fetch, clone,
        access-token:id:..., async, protocol:2 or refs.
    mesh_execution_identifier (str): Mesh execution identifier, only present for
        gRPC/Mesh log lines.
    protocol (str): protocol used, such as http, https, ssh or grpc.
    recorded_time (dfdatetime.DateTimeValues): date and time the log entry
        was recorded.
    remote_address (str): remote IP address(es), including X-Forwarded-For
        proxies. Multiple addresses are comma-separated.
    request_identifier (str): unique request identifier, correlatable with audit
        log.
    request_time (int): time taken to process the request in milliseconds
        (HTTP/SSH) or nanoseconds (gRPC/Mesh), or None if not available.
    session_identifier (str): session identifier.
    ssh_repository_path (str): SSH repository path for SSH requests, or None.
    user_name (str): the name of the authenticated user.
  """

  DATA_TYPE = 'atlassian:bitbucket:access'

  def __init__(self):
    """Initializes event data."""
    super().__init__(data_type=self.DATA_TYPE)
    self.http_request_method = None
    self.http_request_uri = None
    self.http_request_user_agent = None
    self.http_response_bytes_read = None
    self.http_response_bytes_written = None
    self.http_response_code = None
    self.http_version = None
    self.labels = None
    self.mesh_execution_identifier = None
    self.protocol = None
    self.recorded_time = None
    self.remote_address = None
    self.request_identifier = None
    self.request_time = None
    self.session_identifier = None
    self.ssh_repository_path = None
    self.user_name = None


class BitbucketAccessTextPlugin(interface.TextPlugin):
  """Text parser plugin for Atlassian Bitbucket access log files."""

  NAME = 'bitbucket_access'
  DATA_FORMAT = (
      'Atlassian Bitbucket access log (atlassian-bitbucket-access.log) file')

  ENCODING = 'utf-8'

  _INTEGER = pyparsing.Word(pyparsing.nums).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _THREE_DIGITS = pyparsing.Word(pyparsing.nums, exact=3).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  _FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).set_parse_action(
      lambda tokens: int(tokens[0], 10))

  # Pipe as field separator.
  _SEPARATOR = pyparsing.Suppress(pyparsing.Literal('|'))

  # Integer or dash (used for optional numeric fields).
  _INT_OR_DASH = pyparsing.Literal('-') | _INTEGER

  # Remote address: IPv4, IPv6, or comma-separated list of addresses.
  _REMOTE_ADDRESS = pyparsing.Combine(
      pyparsing.Word(pyparsing.alphanums + '.:,')).set_results_name(
          'remote_address')

  # Request identifier: alphanumeric token with @/* prefix and x separators,
  # or '-'.
  _REQUEST_IDENTIFIER = (
      pyparsing.Word(pyparsing.alphanums + '@*-_x') |
      pyparsing.Literal('-'))

  # User name: alphanumeric with common separators, or '-'.
  _USER_NAME = (
      pyparsing.Word(pyparsing.alphanums + '-_./') |
      pyparsing.Literal('-')).set_results_name('user_name')

  # Session identifier: short alphanumeric token or '-'.
  _SESSION_IDENTIFIER = (
      pyparsing.Word(pyparsing.alphanums + '-_') |
      pyparsing.Literal('-'))

  # Date and time format: 2020-09-08 07:53:45,084
  _DATE_TIME = (
      _FOUR_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + pyparsing.Suppress('-') +
      _TWO_DIGITS + _TWO_DIGITS +
      pyparsing.Suppress(':') + _TWO_DIGITS +
      pyparsing.Suppress(':') + _TWO_DIGITS +
      pyparsing.Suppress(',') + _THREE_DIGITS).set_results_name('date_time')

  # HTTP request methods.
  _HTTP_METHODS = [
      'CONNECT', 'DELETE', 'GET', 'HEAD', 'OPTIONS', 'PATCH', 'POST', 'PUT',
      'TRACE']

  # Request URI characters.
  _REQUEST_URI = pyparsing.Word(pyparsing.alphanums + '/-_.?=%&:+<>#~[]@!,()')

  # HTTP request: "METHOD /uri HTTP/1.1"
  _HTTP_REQUEST = pyparsing.Group(
      pyparsing.Suppress('"') +
      pyparsing.one_of(_HTTP_METHODS).set_results_name('http_method') +
      _REQUEST_URI.set_results_name('request_url') +
      pyparsing.Word(pyparsing.alphanums + '/.').set_results_name(
          'http_version') +
      pyparsing.Suppress('"')).set_results_name('http_request')

  # SSH request: "SSH - git-upload-pack '/repo.git'"
  _SSH_REQUEST = pyparsing.Group(
      pyparsing.Suppress('"') +
      pyparsing.Literal('SSH').set_results_name('http_method') +
      pyparsing.Suppress(pyparsing.Literal('-')) +
      pyparsing.Word(pyparsing.alphanums + '-_').set_results_name(
          'request_url') +
      pyparsing.QuotedString("'").set_results_name('ssh_repo') +
      pyparsing.Suppress('"')).set_results_name('http_request')

  # gRPC action: "ServiceName/MethodName"
  _GRPC_REQUEST = pyparsing.Group(
      pyparsing.Suppress('"') +
      pyparsing.SkipTo('"').set_results_name('request_url') +
      pyparsing.Suppress('"')).set_results_name('http_request')

  # Referer and user agent fields: "" "user-agent-string"
  _REFERER = (
      pyparsing.Suppress('"') +
      pyparsing.Optional(pyparsing.Word(
          pyparsing.alphanums + '/-_.?=%&:+<>#~[]@!,()')).set_results_name(
              'referer') +
      pyparsing.Suppress('"'))

  _USER_AGENT = (
      pyparsing.Suppress('"') +
      pyparsing.SkipTo('"').set_results_name('user_agent') +
      pyparsing.Suppress('"'))

  # Labels field: arbitrary text up to the next pipe separator.
  _LABELS = pyparsing.SkipTo(pyparsing.Literal('|')).set_results_name('labels')

  # HTTP/SSH access log line:
  # ip_address | protocol | request_identifier | user | timestamp | "request" |
  # "referer" "user-agent" | status | bytes_read | bytes_written | labels |
  # request_time | session_identifier |
  _HTTP_ACCESS_LOG_LINE = (
      _REMOTE_ADDRESS + _SEPARATOR +
      pyparsing.Word(pyparsing.alphanums).set_results_name('protocol') +
      _SEPARATOR +
      _REQUEST_IDENTIFIER.set_results_name('request_identifier') + _SEPARATOR +
      _USER_NAME + _SEPARATOR +
      _DATE_TIME + _SEPARATOR +
      (_HTTP_REQUEST | _SSH_REQUEST | _GRPC_REQUEST) +
      _SEPARATOR +
      _REFERER + _USER_AGENT + _SEPARATOR +
      _INT_OR_DASH.set_results_name('status_code') + _SEPARATOR +
      _INT_OR_DASH.set_results_name('bytes_read') + _SEPARATOR +
      _INT_OR_DASH.set_results_name('bytes_written') + _SEPARATOR +
      _LABELS + _SEPARATOR +
      _INT_OR_DASH.set_results_name('request_time') + _SEPARATOR +
      _SESSION_IDENTIFIER.set_results_name('session_identifier') + _SEPARATOR +
      pyparsing.Suppress(pyparsing.LineEnd()))

  # gRPC/Mesh access log line:
  # ip_address | grpc | request_identifier | mesh_execution_identifier | user |
  # timestamp | "action" | - | status | bytes_read | bytes_written | mesh_in |
  # mesh_out | duration_ns | session_identifier |
  _GRPC_ACCESS_LOG_LINE = (
      _REMOTE_ADDRESS + _SEPARATOR +
      pyparsing.Literal('grpc').set_results_name('protocol') + _SEPARATOR +
      _REQUEST_IDENTIFIER.set_results_name('request_identifier') + _SEPARATOR +
      (pyparsing.Word(pyparsing.alphanums + '@*-_x') |
       pyparsing.Literal('-')).set_results_name('mesh_execution_identifier') +
      _SEPARATOR +
      _USER_NAME + _SEPARATOR +
      _DATE_TIME + _SEPARATOR +
      _GRPC_REQUEST + _SEPARATOR +
      pyparsing.Suppress(pyparsing.Literal('-')) + _SEPARATOR +
      _INT_OR_DASH.set_results_name('status_code') + _SEPARATOR +
      _INT_OR_DASH.set_results_name('bytes_read') + _SEPARATOR +
      _INT_OR_DASH.set_results_name('bytes_written') + _SEPARATOR +
      _INT_OR_DASH.set_results_name('mesh_in') + _SEPARATOR +
      _INT_OR_DASH.set_results_name('mesh_out') + _SEPARATOR +
      _INT_OR_DASH.set_results_name('duration_ns') + _SEPARATOR +
      _SESSION_IDENTIFIER.set_results_name('session_identifier') + _SEPARATOR +
      pyparsing.Suppress(pyparsing.LineEnd()))

  _LINE_STRUCTURES = [
      ('grpc_access_log', _GRPC_ACCESS_LOG_LINE),
      ('http_access_log', _HTTP_ACCESS_LOG_LINE)]

  VERIFICATION_GRAMMAR = _GRPC_ACCESS_LOG_LINE | _HTTP_ACCESS_LOG_LINE

  VERIFICATION_LITERALS = [
      ' | grpc | ', ' | http | ', ' | https | ', ' | ssh | ']

  def _GetStrippedValue(self, structure, name, default_value=None):
    """Retrieves a token value from a Pyparsing structure and strips '' or '-'.

    Args:
      structure (pyparsing.ParseResults): tokens from a parsed log line.
      name (str): name of the token.
      default_value (Optional[object]): default value.

    Returns:
      object: value in the token or default value if the token is not available
          in the structure.
    """
    value = self._GetValueFromStructure(structure, name)
    if value in (None, '', '-'):
      return default_value

    return value

  def _ParseRecord(self, parser_mediator, key, structure):
    """Parses a pyparsing structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Raises:
      ParseError: if the structure cannot be parsed.
    """
    if key not in ('grpc_access_log', 'http_access_log'):
      raise errors.ParseError(f'Unsupported structure: {key:s}')

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    http_request = self._GetValueFromStructure(structure, 'http_request')

    labels = self._GetValueFromStructure(
        structure, 'labels', default_value='').strip()

    event_data = BitbucketAccessEventData()
    event_data.http_response_bytes_read = self._GetStrippedValue(
        structure, 'bytes_read')
    event_data.http_response_bytes_written = self._GetStrippedValue(
        structure, 'bytes_written')
    event_data.http_response_code = self._GetStrippedValue(
        structure, 'status_code')
    event_data.http_request_user_agent = self._GetStringValueFromStructure(
        structure, 'user_agent') or None
    event_data.labels = None if not labels or labels == '-' else labels
    event_data.mesh_execution_identifier = self._GetStrippedValue(
        structure, 'mesh_execution_identifier')
    event_data.protocol = self._GetValueFromStructure(structure, 'protocol')
    event_data.recorded_time = self._ParseTimeElements(time_elements_structure)
    event_data.remote_address = self._GetValueFromStructure(
        structure, 'remote_address')
    event_data.request_identifier = self._GetStrippedValue(
        structure, 'request_identifier')
    event_data.request_time = self._GetStrippedValue(structure, 'request_time')
    event_data.session_identifier = self._GetStrippedValue(
        structure, 'session_identifier')
    event_data.user_name = self._GetStrippedValue(structure, 'user_name')

    if http_request:
      event_data.http_request_method = self._GetValueFromStructure(
          http_request, 'http_method')
      event_data.http_request_uri = self._GetValueFromStructure(
          http_request, 'request_url')
      event_data.http_version = self._GetValueFromStructure(
          http_request, 'http_version') or None
      event_data.ssh_repository_path = self._GetValueFromStructure(
          http_request, 'ssh_repo') or None

    parser_mediator.ProduceEventData(event_data)

  def _ParseTimeElements(self, time_elements_structure):
    """Parses date and time elements of a log line.

    Args:
      time_elements_structure (pyparsing.ParseResults): date and time elements.

    Returns:
      dfdatetime.TimeElements: date and time value.

    Raises:
      ParseError: if a valid date and time value cannot be derived.
    """
    try:
      date_time = dfdatetime_time_elements.TimeElementsInMilliseconds(
          time_elements_tuple=time_elements_structure)

      date_time.is_local_time = True

      return date_time

    except (IndexError, TypeError, ValueError) as exception:
      raise errors.ParseError(
          f'Unable to parse time elements with error: {exception!s}')

  def CheckRequiredFormat(self, parser_mediator, text_reader):
    """Check if the log record has the minimal structure required by the plugin.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfVFS.
      text_reader (EncodedTextReader): text reader.

    Returns:
      bool: True if this is the correct plugin, False otherwise.
    """
    try:
      structure = self._VerifyString(text_reader.lines)
    except errors.ParseError:
      return False

    time_elements_structure = self._GetValueFromStructure(
        structure, 'date_time')

    try:
      self._ParseTimeElements(time_elements_structure)
    except errors.ParseError:
      return False

    return True


text_parser.TextLogParser.RegisterPlugin(BitbucketAccessTextPlugin)
